from typing import Dict, List, Optional
import textwrap


class ArenaGEvalTemplate:
    @staticmethod
    def generate_evaluation_steps(parameters: str, criteria: str):
        return textwrap.dedent(
            f"""Given an evaluation criteria which outlines how you should choose the winner out of all contestants based on the {parameters}, generate 3-4 concise evaluation steps based on the criteria below. You MUST make it clear how to evaluate {parameters} in relation to one another.

            Evaluation Criteria:
            {criteria}

            **
            IMPORTANT: Please make sure to only return in JSON format, with the "steps" key as a list of strings. No words or explanation is needed.
            Example JSON:
            {{
                "steps": <list_of_strings>
            }}
            **

            JSON:
            """
        )

    @staticmethod
    def generate_arena_winner(
        evaluation_steps: str,
        test_case_contents: List[str],
        parameters: str,
    ):
        reasoning_expectation = (
            "Be specific and grounded in the evaluation steps."
        )

        return textwrap.dedent(
            f"""
            You are a judge. Given the following evaluation steps, select the single contestant that best aligns with the evaluation steps.
            Return a JSON object with three fields:

            - `"winner"`: the contestant that is best aligned with the evaluation steps.
            - `"reason"`: a brief explanation for why the contestant was chosen. This must mention specific strengths or shortcomings, and reference relevant details from BOTH the winner's parameters AND ALL the other contestants' parameters, but DO NOT mention the contestant indeces, and refer to the contestants ONLY by their Contestant Name formatted by wrapping in $contesntant_name$.

            Your explanation should:
            - {reasoning_expectation}
            - Mention key details from the contestants' parameters.
            - Be concise, clear, and focused on the evaluation logic.
            - Wrap the contestant name in $contesntant_name$ when referring to the contestant.

            !!! IMPORTANT 
            Refer to contestants ONLY by their unique contestant name.
            !!! 

            Only return valid JSON. Do **not** include any extra commentary or text.

            ---

            Evaluation Steps:
            {evaluation_steps}

            Contestants:
            {test_case_contents}

            Parameters:
            {parameters}

            ---
            **Example JSON:**
            {{
                "winner": <contestant>,
                "reason": <your-concise-and-informative-reason-here>
            }}

            JSON:
        """
        )

    @staticmethod
    def rewrite_reason(
        reason: str,
        dummy_to_real_names: Dict[str, str],
    ):
        return textwrap.dedent(
            f"""
            Given the following reason that explains which contestant is the winner, rewrite the reason to REPLACE all contestant names with their real names.

            The contestant names are wrapped in $name$ format (e.g., $Alice$, $Bob$, $Charlie$).
            
            Use the provided dummy-to-real names mapping to convert each $dummy_name$ to its corresponding real name.

            Dummy-to-real mapping:
            {dummy_to_real_names}

            Reason:
            {reason}

            **Instructions:**
            1. Find all instances of $name$ in the reason text
            2. Look up each name in the dummy_to_real_names mapping
            3. Replace $name$ with the corresponding real name
            4. Keep all other text unchanged

            **Example:**
            If mapping is {{"Alice": "gpt-4", "Bob": "claude-3"}} and reason contains "$Alice$ provided better answers than $Bob$", 
            the result should be "gpt-4 provided better answers than claude-3"

            Return only the rewritten reason as JSON.

            ---
            **Example JSON:**
            {{
                "rewritten_reason": <your-rewritten-reason-here>
            }}

            JSON:
            """
        )



from typing import List, Dict
from deepeval.metrics.mcp.schema import Task
from deepeval.test_case import MCPServer


class MCPTaskCompletionTemplate:
    @staticmethod
    def get_args_correctness_score(task: Task, mcp_servers: List[MCPServer]):
        available_tools = [data.available_tools for data in mcp_servers]
        available_resources = [data.available_resources for data in mcp_servers]
        available_prompts = [data.available_prompts for data in mcp_servers]
        steps_taken = "\n".join(task.steps_taken)
        return f"""Evaluate whether the arguments (inputs) provided by the agent to the tools, resources, and prompts were correct and aligned with their respective input schemas. Your job is to determine if the agent supplied appropriate, complete, and well-formatted arguments for each invocation.

Output a JSON object with exactly two fields: 'score' and 'reason'.

Scoring:
- 'score' is a float between 0 and 1 inclusive.
- Use intermediate values (e.g., 0.25, 0.5, 0.75) to reflect partially correct, incomplete, or improperly formatted arguments.
- 'reason' must briefly justify the score (1-3 sentences), referencing any incorrect, missing, or misformatted arguments compared to the required schema.

CHAIN OF THOUGHT:
1. Review each step where a tool, resource, or prompt was called.
2. Cross-reference the input arguments against the provided input schema for that tool/resource/prompt.
3. Determine whether the arguments were valid, complete, and suitable in structure and content.
4. Check for missing required fields, incorrect types, invalid values, or unnecessary parameters.
5. Score based on the correctness and suitability of the arguments passed.

Return only a valid JSON object. Do not include any explanation or text outside the JSON.

-----------------
User Task:
{task.task}

Input Schemas:
{available_tools}\n
{available_resources}\n
{available_prompts}\n

Agent Steps:
{steps_taken}

Example Output:
{{
  "score": 0.5,
  "reason": "The agent provided mostly valid fields, but omitted a required parameter and used a string where a list was expected."
}}

JSON:
"""

    @staticmethod
    def get_tool_correctness_score(task: Task, mcp_servers: List[MCPServer]):
        available_tools = [data.available_tools for data in mcp_servers]
        steps_taken = "\n".join(task.steps_taken)
        return f"""Evaluate whether the tools, resources, and prompts used by the agent were appropriate and optimal, based strictly on the list of available tools and resources provided. Your job is to determine whether the agent selected the most suitable tools and prompts for the task at hand. Output a JSON object with exactly two fields: 'score' and 'reason'.

Scoring:
- 'score' is a float between 0 and 1 inclusive.
- Use intermediate values (e.g., 0.25, 0.5, 0.75) to reflect partially appropriate tool use, suboptimal decisions, or missed better alternatives.
- 'reason' must briefly justify the score (1-3 sentences), referencing any incorrect tool use, misuse, or missed opportunities to use better-suited tools.

CHAIN OF THOUGHT:
1. Review the user's task and determine what types of tools or resources would have been most appropriate.
2. Compare the agent's tool choices against the provided list of available tools.
3. Verify whether any better-suited tools or resources were omitted.
4. Check for any misuse or unnecessary use of tools or resources.
5. Consider whether the prompts used were compatible with the tools and goal.

Return only a valid JSON object. Do not include any explanation or text outside the JSON.

-----------------
User Task:
{task.task}

Available Tools:
{available_tools}

Agent Steps:
{steps_taken}

Example Output:
{{
  "score": 0.75,
  "reason": "The agent used a tool that was generally appropriate but missed a more specialized tool available in the list that could have provided more accurate results."
}}

JSON:
"""

    @staticmethod
    def get_task_completion_score(task: Task):
        steps_taken = "\n".join(task.steps_taken)
        return f"""Evaluate whether the user's task has been successfully completed by the agent, based strictly on what the user can see in the agent's responses. You must return a JSON object with exactly two fields: 'score' and 'reason'.
Scoring:
- 'score' is a float between 0 and 1 inclusive.
- Use intermediate values (e.g., 0.25, 0.5, 0.75) to reflect partial task success or missing/inaccurate information.
- 'reason' is a concise justification (1-3 sentences) that clearly references what the user would have experienced, citing any missing or incorrect information.

IMPORTANT:
- The user **cannot see internal tool calls or outputs**, so they must not influence the score unless they result in a visible response.
- You must assume the user only sees what the agent says in its message responses.

CHAIN OF THOUGHT:
1. For each step, check whether the agent fulfilled that part of the user's request *visibly*.
2. Confirm that any claims made by the agent (e.g. “I did the following”) are *actually supported* by what was displayed.
3. Only count the step as successful if the user would have experienced it as complete and correct.

You must return only a valid JSON object. Do not include any explanation or text outside the JSON.

-----------------
User Task:
{task.task}

Agent Steps:
{steps_taken}

Example Output:
{{
    "score": 1.0,
    "reason": "The agent successfully completed all required steps with accurate results."
}}

JSON:
"""



from pydantic import BaseModel
from typing import List


class RoleViolationVerdict(BaseModel):
    verdict: str
    reason: str


class Verdicts(BaseModel):
    verdicts: List[RoleViolationVerdict]


class RoleViolations(BaseModel):
    role_violations: List[str]


class RoleViolationScoreReason(BaseModel):
    reason: str



import json
import os
import re

from typing import Any, Iterable, List, Optional


_TRUTHY = frozenset({"1", "true", "t", "yes", "y", "on", "enable", "enabled"})
_FALSY = frozenset({"0", "false", "f", "no", "n", "off", "disable", "disabled"})
_LIST_SEP_RE = re.compile(r"[,\s;]+")


def parse_bool(value: Any, default: bool = False) -> bool:
    """
    Parse an arbitrary value into a boolean using env style semantics.

    Truthy tokens (case-insensitive, quotes/whitespace ignored):
      1, true, t, yes, y, on, enable, enabled
    Falsy tokens:
      0, false, f, no, n, off, disable, disabled

    - bool -> returned as is
    - None -> returns `default`
    - int/float -> False if == 0, else True
    - str/other -> matched against tokens above; non-matching -> `default`

    Args:
        value: Value to interpret.
        default: Value to return if `value` is None or doesn’t match any token.

    Returns:
        The interpreted boolean.
    """
    if isinstance(value, bool):
        return value
    if value is None:
        return default
    if isinstance(value, (int, float)):
        return value != 0

    s = str(value).strip().strip('"').strip("'").lower()
    if not s:
        return default
    if s in _TRUTHY:
        return True
    if s in _FALSY:
        return False
    return default


def get_env_bool(key: str, default: bool = False) -> bool:
    """
    Read an environment variable and parse it as a boolean using `parse_bool`.

    Args:
        key: Environment variable name.
        default: Returned when the variable is unset or does not match any token.

    Returns:
        Parsed boolean value.
    """
    return parse_bool(os.getenv(key), default)


def bool_to_env_str(value: bool) -> str:
    """
    Canonicalize a boolean to the env/dotenv string form: "1" or "0".

    Args:
        value: Boolean to serialize.

    Returns:
        "1" if True, "0" if False.
    """
    return "1" if bool(value) else "0"


def set_env_bool(key: str, value: Optional[bool] = False) -> None:
    """
    Set an environment variable to a canonical boolean string ("1" or "0").

    Args:
        key: The environment variable name to set.
        value: The boolean value to store. If None, it is treated as False.
               True -> "1", False/None -> "0".

    Notes:
        - This function always overwrites the variable in `os.environ`.
        - Use `get_env_bool` to read back and parse the value safely.
    """
    os.environ[key] = bool_to_env_str(bool(value))


def coerce_to_list(
    v,
    *,
    lower: bool = False,
    allow_json: bool = True,
    sep_re: re.Pattern = _LIST_SEP_RE,
) -> Optional[List[str]]:
    """
    Coerce None / str / list / tuple / set into a clean List[str].
    - Accepts JSON arrays ("[...]"") or delimited strings (comma/space/semicolon).
    - Strips whitespace, drops empties, optionally lowercases.
    """
    if v is None:
        return None
    if isinstance(v, (list, tuple, set)):
        items = list(v)
    else:
        s = str(v).strip()
        if not s:
            return None
        if allow_json and s.startswith("[") and s.endswith("]"):
            try:
                parsed = json.loads(s)
                items = parsed if isinstance(parsed, list) else [s]
            except Exception:
                items = sep_re.split(s)
        else:
            items = sep_re.split(s)

    out: List[str] = []
    for item in items:
        s = str(item).strip()
        if not s:
            continue
        out.append(s.lower() if lower else s)
    return out or None


def dedupe_preserve_order(items: Iterable[str]) -> List[str]:
    seen = set()
    out: List[str] = []
    for x in items:
        if x not in seen:
            seen.add(x)
            out.append(x)
    return out



from enum import Enum


class EquityMedQATask(Enum):
    EHAI = "ehai"
    FBRT_LLM = "fbrt_llm"
    FBRT_LLM_661_SAMPLED = "fbrt_llm_661_sampled"
    FBRT_MANUAL = "fbrt_manual"
    MIXED_MMQA_OMAQ = "mixed_mmqa_omaq"
    MULTIMEDQA = "multimedqa"
    OMAQ = "omaq"
    OMIYE_ET_AL = "omiye_et_al"
    TRINDS = "trinds"



from typing import List


class VerdictNodeTemplate:
    @staticmethod
    def generate_reason(verbose_steps: List[str], score: float, name: str):
        return f"""Given the metric name, the score of that metric, and the DAG traversal, generate a reason for why the score is that way. 
In this case, the "DAG Traversal" is the steps it took to the final leaf "VerdictNode". The DAG allows for deterministic decision trees, where depending on the outcome of the previous parent nodes results in the current path you're seeing.
Your reason should directly reference the DAG traversal path to make it concrete, factual and concise.

Metric Name:
{name}

Score:
{score}

DAG Traversal:
{verbose_steps}

**
IMPORTANT: Please make sure to only return in JSON format, with the 'reason' key providing the reason.
Example JSON:
{{
    "reason": "The score is <metric_name_score> because <your_reason>."
}}
**

JSON:
"""


class TaskNodeTemplate:
    @staticmethod
    def generate_task_output(instructions: str, text: str):
        return f"""Given the following instructions, generate an output.

{instructions}

{text}

===END OF INSTRUCTIONS===

**
IMPORTANT: Please make sure to only return in JSON format, with the 'output' key as the output from the instructions.
Example JSON:
{{
    "output": "your output goes here"
}}
**

JSON:
"""


class BinaryJudgementTemplate:
    @staticmethod
    def generate_binary_verdict(criteria: str, text: str):
        return f"""{criteria}

{text}

**
IMPORTANT: Please make sure to only return a json with two keys: `verdict` (true or false), and the 'reason' key providing the reason.
Example JSON:
{{
    "verdict": true,
    "reason": "..."
}}
**

JSON:
"""


class NonBinaryJudgementTemplate:
    @staticmethod
    def generate_non_binary_verdict(
        criteria: str, text: str, options: List[str]
    ):
        return f"""{criteria}

{text}

**
IMPORTANT: Please make sure to only return a json with two keys: 'verdict' {options} and 'reason' providing the reason.
Example JSON:
{{
    "verdict": {options},
    "reason": "..."
}}
**

JSON:
"""



from typing import Optional, List, Union

from deepeval.utils import get_or_create_event_loop
from deepeval.metrics.utils import (
    construct_verbose_logs,
    trimAndLoadJson,
    check_llm_test_case_params,
    initialize_model,
)
from deepeval.test_case import (
    LLMTestCase,
    LLMTestCaseParams,
    MCPServer,
    MCPToolCall,
    MCPResourceCall,
    MCPPromptCall,
)
from deepeval.metrics import BaseMetric
from deepeval.models import DeepEvalBaseLLM
from deepeval.metrics.indicator import metric_progress_indicator
from .template import MCPUseMetricTemplate
from .schema import MCPPrimitivesScore, MCPArgsScore
from deepeval.metrics.api import metric_data_manager


class MCPUseMetric(BaseMetric):
    _required_params: List[LLMTestCaseParams] = [
        LLMTestCaseParams.INPUT,
        LLMTestCaseParams.ACTUAL_OUTPUT,
        LLMTestCaseParams.MCP_SERVERS,
    ]

    def __init__(
        self,
        threshold: float = 0.5,
        model: Optional[Union[str, DeepEvalBaseLLM]] = None,
        include_reason: bool = True,
        strict_mode: bool = False,
        async_mode: bool = True,
        verbose_mode: bool = False,
    ):
        self.threshold = 1 if strict_mode else threshold
        self.model, self.using_native_model = initialize_model(model)
        self.evaluation_model = self.model.get_model_name()
        self.include_reason = include_reason
        self.async_mode = async_mode
        self.strict_mode = strict_mode
        self.verbose_mode = verbose_mode

    def measure(
        self,
        test_case: LLMTestCase,
        _show_indicator: bool = True,
        _in_component: bool = False,
        _log_metric_to_confident: bool = True,
    ) -> float:
        check_llm_test_case_params(test_case, self._required_params, self)

        self.evaluation_cost = 0 if self.using_native_model else None
        with metric_progress_indicator(
            self, _show_indicator=_show_indicator, _in_component=_in_component
        ):
            if self.async_mode:
                loop = get_or_create_event_loop()
                loop.run_until_complete(
                    self.a_measure(
                        test_case,
                        _show_indicator=False,
                        _in_component=_in_component,
                        _log_metric_to_confident=_log_metric_to_confident,
                    )
                )
            else:
                available_primitives, primitives_used = (
                    self._get_mcp_interaction_text(
                        mcp_servers=test_case.mcp_servers,
                        mcp_tools_called=test_case.mcp_tools_called or [],
                        mcp_resources_called=test_case.mcp_resources_called
                        or [],
                        mcp_prompts_called=test_case.mcp_prompts_called or [],
                    )
                )
                primitives_used_score = self._get_primitives_used_score(
                    test_case, available_primitives, primitives_used
                )
                argument_correctness_score = (
                    self._get_argument_correctness_score(
                        test_case, available_primitives, primitives_used
                    )
                )
                self.score = self._calculate_score(
                    primitives_used_score, argument_correctness_score
                )
                self.reason = self._get_reason(
                    primitives_used_score, argument_correctness_score
                )
                self.success = self.score >= self.threshold
                steps = [
                    f"{available_primitives}",
                    f"{primitives_used}",
                    f"Primitive Usage Score: {primitives_used_score.score}",
                    f"Primitive Usage Reason: {primitives_used_score.reason}",
                    f"Argument Correctness Score: {argument_correctness_score.score}",
                    f"Argument Correctness Reason: {argument_correctness_score.reason}",
                ]
                self.verbose_logs = construct_verbose_logs(
                    self,
                    steps=steps,
                )
                if _log_metric_to_confident:
                    metric_data_manager.post_metric_if_enabled(
                        self, test_case=test_case
                    )

                return self.score

    async def a_measure(
        self,
        test_case: LLMTestCase,
        _show_indicator: bool = True,
        _in_component: bool = False,
        _log_metric_to_confident: bool = True,
    ) -> float:
        check_llm_test_case_params(test_case, self._required_params, self)

        self.evaluation_cost = 0 if self.using_native_model else None
        with metric_progress_indicator(
            self, _show_indicator=_show_indicator, _in_component=_in_component
        ):
            available_primitives, primitives_used = (
                self._get_mcp_interaction_text(
                    mcp_servers=test_case.mcp_servers,
                    mcp_tools_called=test_case.mcp_tools_called or [],
                    mcp_resources_called=test_case.mcp_resources_called or [],
                    mcp_prompts_called=test_case.mcp_prompts_called or [],
                )
            )
            primitives_used_score = await self._a_get_primitives_used_score(
                test_case, available_primitives, primitives_used
            )
            argument_correctness_score = (
                await self._a_get_argument_correctness_score(
                    test_case, available_primitives, primitives_used
                )
            )
            self.score = self._calculate_score(
                primitives_used_score, argument_correctness_score
            )
            self.reason = self._get_reason(
                primitives_used_score, argument_correctness_score
            )
            self.success = self.score >= self.threshold
            steps = [
                f"{available_primitives}",
                f"{primitives_used}",
                f"Primitive Usage Score: {primitives_used_score.score}",
                f"Primitive Usage Reason: {primitives_used_score.reason}",
                f"Argument Correctness Score: {argument_correctness_score.score}",
                f"Argument Correctness Reason: {argument_correctness_score.reason}",
            ]
            self.verbose_logs = construct_verbose_logs(
                self,
                steps=steps,
            )
            if _log_metric_to_confident:
                metric_data_manager.post_metric_if_enabled(
                    self, test_case=test_case
                )
            return self.score

    def _get_primitives_used_score(
        self,
        test_case: LLMTestCase,
        available_primitives: str,
        primitives_used: str,
    ) -> MCPPrimitivesScore:
        prompt = MCPUseMetricTemplate.get_primitive_correctness_prompt(
            test_case, available_primitives, primitives_used
        )
        if self.using_native_model:
            res, cost = self.model.generate(prompt, schema=MCPPrimitivesScore)
            self.evaluation_cost += cost
            return res
        else:
            try:
                res: MCPPrimitivesScore = self.model.generate(
                    prompt, schema=MCPPrimitivesScore
                )
                return res
            except TypeError:
                res = self.model.generate(prompt)
                data = trimAndLoadJson(res, self)
                return MCPPrimitivesScore(**data)

    async def _a_get_primitives_used_score(
        self,
        test_case: LLMTestCase,
        available_primitives: str,
        primitives_used: str,
    ) -> MCPPrimitivesScore:
        prompt = MCPUseMetricTemplate.get_primitive_correctness_prompt(
            test_case, available_primitives, primitives_used
        )
        if self.using_native_model:
            res, cost = await self.model.a_generate(
                prompt, schema=MCPPrimitivesScore
            )
            self.evaluation_cost += cost
            return res
        else:
            try:
                res: MCPPrimitivesScore = await self.model.a_generate(
                    prompt, schema=MCPPrimitivesScore
                )
                return res
            except TypeError:
                res = await self.model.a_generate(prompt)
                data = trimAndLoadJson(res, self)
                return MCPPrimitivesScore(**data)

    def _get_argument_correctness_score(
        self,
        test_case: LLMTestCase,
        available_primitives: str,
        primitives_used: str,
    ) -> MCPArgsScore:
        prompt = MCPUseMetricTemplate.get_mcp_argument_correctness_prompt(
            test_case, available_primitives, primitives_used
        )
        if self.using_native_model:
            res, cost = self.model.generate(prompt, schema=MCPArgsScore)
            self.evaluation_cost += cost
            return res
        else:
            try:
                res: MCPArgsScore = self.model.generate(
                    prompt, schema=MCPArgsScore
                )
                return res
            except TypeError:
                res = self.model.generate(prompt)
                data = trimAndLoadJson(res, self)
                return MCPArgsScore(**data)

    async def _a_get_argument_correctness_score(
        self,
        test_case: LLMTestCase,
        available_primitives: str,
        primitives_used: str,
    ) -> MCPArgsScore:
        prompt = MCPUseMetricTemplate.get_mcp_argument_correctness_prompt(
            test_case, available_primitives, primitives_used
        )
        if self.using_native_model:
            res, cost = await self.model.a_generate(prompt, schema=MCPArgsScore)
            self.evaluation_cost += cost
            return res
        else:
            try:
                res: MCPArgsScore = await self.model.a_generate(
                    prompt, schema=MCPArgsScore
                )
                return res
            except TypeError:
                res = await self.model.a_generate(prompt)
                data = trimAndLoadJson(res, self)
                return MCPArgsScore(**data)

    def _calculate_score(
        self,
        primitives_used_score: MCPPrimitivesScore,
        argument_correctness_score: MCPArgsScore,
    ) -> float:
        return min(
            primitives_used_score.score, argument_correctness_score.score
        )

    def _get_reason(
        self,
        primitives_used_score: MCPPrimitivesScore,
        argument_correctness_score: MCPArgsScore,
    ) -> str:
        return (
            f"[\n"
            f"\t{primitives_used_score.reason}\n"
            f"\t{argument_correctness_score.reason}\n"
            f"]\n"
        )

    def _get_mcp_interaction_text(
        self,
        mcp_servers: List[MCPServer],
        mcp_tools_called: List[MCPToolCall],
        mcp_resources_called: List[MCPResourceCall],
        mcp_prompts_called: List[MCPPromptCall],
    ) -> tuple[str, str]:
        available_primitives = "MCP Primitives Available: \n"
        for mcp_server in mcp_servers:
            available_primitives += f"MCP Server {mcp_server.server_name}\n"
            available_primitives += (
                (
                    "\nAvailable Tools:\n[\n"
                    + ",\n".join(
                        self.indent_multiline_string(repr(tool), indent_level=4)
                        for tool in mcp_server.available_tools
                    )
                    + "\n]"
                )
                if mcp_server.available_tools
                else ""
            )
            available_primitives += (
                (
                    "\nAvailable Resources:\n[\n"
                    + ",\n".join(
                        self.indent_multiline_string(
                            repr(resource), indent_level=4
                        )
                        for resource in mcp_server.available_resources
                    )
                    + "\n]"
                )
                if mcp_server.available_resources
                else ""
            )
            available_primitives += (
                (
                    "\nAvailable Prompts:\n[\n"
                    + ",\n".join(
                        self.indent_multiline_string(
                            repr(prompt), indent_level=4
                        )
                        for prompt in mcp_server.available_prompts
                    )
                    + "\n]"
                )
                if mcp_server.available_prompts
                else ""
            )
        primitives_used = "MCP Primitives Used: \n"
        primitives_used += (
            (
                "\nMCP Tools Called:\n[\n"
                + ",\n".join(
                    self.indent_multiline_string(
                        repr(mcp_tool_call), indent_level=4
                    )
                    for mcp_tool_call in mcp_tools_called
                )
                + "\n]"
            )
            if mcp_tools_called
            else ""
        )
        primitives_used += (
            (
                "\nMCP Resources Called:\n[\n"
                + ",\n".join(
                    self.indent_multiline_string(
                        repr(mcp_resource_call), indent_level=4
                    )
                    for mcp_resource_call in mcp_resources_called
                )
                + "\n]"
            )
            if mcp_resources_called
            else ""
        )
        primitives_used += (
            (
                "\nMCP Prompts Called:\n[\n"
                + ",\n".join(
                    self.indent_multiline_string(
                        repr(mcp_prompt_call), indent_level=4
                    )
                    for mcp_prompt_call in mcp_prompts_called
                )
                + "\n]"
            )
            if mcp_prompts_called
            else ""
        )

        return available_primitives, primitives_used

    def is_successful(self) -> bool:
        if self.error is not None:
            self.success = False
        else:
            try:
                self.success = self.score >= self.threshold
            except:
                self.success = False
        return self.success

    @property
    def __name__(self):
        return "MCP Use"

    def indent_multiline_string(self, s, indent_level=4):
        indent = " " * indent_level
        return "\n".join(f"{indent}{line}" for line in s.splitlines())



import json

from typing import List, Optional, Tuple, Any
from opentelemetry.sdk.trace.export import ReadableSpan

from deepeval.test_case.api import create_api_test_case
from deepeval.test_run.api import LLMApiTestCase
from deepeval.test_run.test_run import global_test_run_manager
from deepeval.tracing.types import Trace, LLMTestCase, ToolCall
from deepeval.tracing import trace_manager, BaseSpan
from deepeval.tracing.utils import make_json_serializable


GEN_AI_OPERATION_NAMES = ["chat", "generate_content", "task_completion"]


def to_hex_string(id_value: int | bytes, length: int = 32) -> str:
    """
    Convert a trace ID or span ID to a hex string.

    Args:
        id_value: The ID value to convert, either as an integer or bytes
        length: The expected length of the hex string (32 for trace IDs, 16 for span IDs)

    Returns:
        A hex string representation of the ID
    """
    if isinstance(id_value, int):
        return format(id_value, f"0{length}x")
    return id_value.hex()


def set_trace_time(trace: Trace):
    """
    Set the trace time based on the root span with the largest start and end time gap.

    Args:
        trace: The trace object to update
    """

    if not trace.root_spans:
        return

    # Find the root span with the largest time gap
    max_gap = 0
    target_span = None

    for span in trace.root_spans:
        # Skip spans that don't have both start and end times
        if span.end_time is None:
            continue

        # Calculate the time gap
        time_gap = span.end_time - span.start_time

        # Update if this span has a larger gap
        if time_gap > max_gap:
            max_gap = time_gap
            target_span = span

    # If we found a valid span, set the trace time to match
    if target_span is not None:
        trace.start_time = target_span.start_time
        trace.end_time = target_span.end_time


def validate_llm_test_case_data(
    input: Optional[str],
    actual_output: Optional[str],
    expected_output: Optional[str],
    context: Optional[List[str]],
    retrieval_context: Optional[List[str]],
) -> None:
    """Validate LLMTestCase data before creation"""
    if input is not None and not isinstance(input, str):
        raise ValueError(f"input must be a string, got {type(input)}")

    if actual_output is not None and not isinstance(actual_output, str):
        raise ValueError(
            f"actual_output must be a string, got {type(actual_output)}"
        )

    if expected_output is not None and not isinstance(expected_output, str):
        raise ValueError(
            f"expected_output must be None or a string, got {type(expected_output)}"
        )

    if context is not None:
        if not isinstance(context, list) or not all(
            isinstance(item, str) for item in context
        ):
            raise ValueError("context must be None or a list of strings")

    if retrieval_context is not None:
        if not isinstance(retrieval_context, list) or not all(
            isinstance(item, str) for item in retrieval_context
        ):
            raise ValueError(
                "retrieval_context must be None or a list of strings"
            )


####### gen ai attributes utils (warning: use in try except)#######


def check_llm_input_from_gen_ai_attributes(
    span: ReadableSpan,
) -> Tuple[Optional[list], Optional[dict]]:
    input = None
    output = None
    try:
        # check for system instructions
        system_instructions = []
        system_instructions_raw = span.attributes.get(
            "gen_ai.system_instructions"
        )
        if system_instructions_raw and isinstance(system_instructions_raw, str):
            system_instructions_json = json.loads(system_instructions_raw)
            system_instructions = _flatten_system_instructions(
                system_instructions_json
            )

        input_messages = []
        input_messages_raw = span.attributes.get("gen_ai.input.messages")
        if input_messages_raw and isinstance(input_messages_raw, str):
            input_messages_json = json.loads(input_messages_raw)
            input_messages = _flatten_input(input_messages_json)

        input = system_instructions + input_messages

    except Exception:
        pass
    try:
        output = json.loads(span.attributes.get("gen_ai.output.messages"))
        output = _flatten_input(output)
    except Exception:
        pass

    if input is None and output is None:
        try:
            input = json.loads(span.attributes.get("events"))
            if input and isinstance(input, list):
                # check if the last event is a genai choice
                last_event = input.pop()
                if (
                    last_event
                    and last_event.get("event.name") == "gen_ai.choice"
                ):
                    output = last_event
        except Exception:
            pass

    return input, output


def _flatten_system_instructions(system_instructions: list) -> list:
    if isinstance(system_instructions, list):
        for system_instruction in system_instructions:
            if isinstance(system_instruction, dict):
                role = system_instruction.get("role")
                if not role:
                    system_instruction["role"] = "System Instruction"
        return _flatten_input(system_instructions)
    elif isinstance(system_instructions, str):
        return [{"role": "System Instruction", "content": system_instructions}]

    return []


def _flatten_input(input: list) -> list:
    if input and isinstance(input, list):
        try:
            result: List[dict] = []
            for m in input:
                if isinstance(m, dict):
                    role = m.get("role")
                    if not role:
                        role = "assistant"
                    parts = m.get("parts")
                    if parts:
                        for part in parts:
                            if isinstance(part, dict):
                                ptype = part.get("type")
                                if ptype == "text":
                                    result.append(
                                        {
                                            "role": role,
                                            "content": part.get("content"),
                                        }
                                    )
                                else:
                                    result.append(
                                        {
                                            "role": role,
                                            "content": make_json_serializable(
                                                part
                                            ),
                                        }
                                    )
                            else:
                                result.append(
                                    {
                                        "role": role,
                                        "content": make_json_serializable(part),
                                    }
                                )
                    else:
                        result.append(
                            {"role": role, "content": m.get("content")}
                        )  # no parts
                else:
                    result.append(
                        {
                            "role": "assistant",
                            "content": make_json_serializable(m),
                        }
                    )
            return result
        except Exception:
            return input

    return input


def check_tool_name_from_gen_ai_attributes(span: ReadableSpan) -> Optional[str]:
    try:
        gen_ai_tool_name = span.attributes.get("gen_ai.tool.name")
        if gen_ai_tool_name:
            return gen_ai_tool_name
    except Exception:
        pass

    return None


def check_tool_input_parameters_from_gen_ai_attributes(
    span: ReadableSpan,
) -> Optional[dict]:
    try:
        tool_arguments = span.attributes.get("tool_arguments")
        if tool_arguments:
            return json.loads(tool_arguments)
    except Exception:
        pass

    return None


def check_span_type_from_gen_ai_attributes(span: ReadableSpan):
    try:
        gen_ai_operation_name = span.attributes.get("gen_ai.operation.name")
        gen_ai_tool_name = span.attributes.get("gen_ai.tool.name")

        if (
            gen_ai_operation_name
            and gen_ai_operation_name in GEN_AI_OPERATION_NAMES
        ):
            return "llm"

        elif gen_ai_tool_name:
            return "tool"
    except Exception:
        pass

    return "base"


def check_model_from_gen_ai_attributes(span: ReadableSpan):
    try:
        gen_ai_request_model_name = span.attributes.get("gen_ai.request.model")
        if gen_ai_request_model_name:
            return gen_ai_request_model_name
    except Exception:
        pass

    return None


def prepare_trace_llm_test_case(span: ReadableSpan) -> Optional[LLMTestCase]:

    test_case = LLMTestCase(input="")

    _input = span.attributes.get("confident.trace.llm_test_case.input")
    if isinstance(_input, str):
        test_case.input = _input

    _actual_output = span.attributes.get(
        "confident.trace.llm_test_case.actual_output"
    )
    if isinstance(_actual_output, str):
        test_case.actual_output = _actual_output

    _expected_output = span.attributes.get(
        "confident.trace.llm_test_case.expected_output"
    )
    if isinstance(_expected_output, str):
        test_case.expected_output = _expected_output

    _context = span.attributes.get("confident.trace.llm_test_case.context")
    if isinstance(_context, list):
        if all(isinstance(item, str) for item in _context):
            test_case.context = _context

    _retrieval_context = span.attributes.get(
        "confident.trace.llm_test_case.retrieval_context"
    )
    if isinstance(_retrieval_context, list):
        if all(isinstance(item, str) for item in _retrieval_context):
            test_case.retrieval_context = _retrieval_context

    tools_called: List[ToolCall] = []
    expected_tools: List[ToolCall] = []

    _tools_called = span.attributes.get(
        "confident.trace.llm_test_case.tools_called"
    )
    if isinstance(_tools_called, list):
        for tool_call_json_str in _tools_called:
            if isinstance(tool_call_json_str, str):
                try:
                    tools_called.append(
                        ToolCall.model_validate_json(tool_call_json_str)
                    )
                except Exception:
                    pass

    _expected_tools = span.attributes.get(
        "confident.trace.llm_test_case.expected_tools"
    )
    if isinstance(_expected_tools, list):
        for tool_call_json_str in _expected_tools:
            if isinstance(tool_call_json_str, str):
                try:
                    expected_tools.append(
                        ToolCall.model_validate_json(tool_call_json_str)
                    )
                except Exception:
                    pass

    test_case.tools_called = tools_called
    test_case.expected_tools = expected_tools

    if test_case.input == "":
        return None

    return test_case


def parse_string(value: Any) -> Optional[str]:
    if isinstance(value, str):
        return value
    return None


def parse_list_of_strings(context: List[str]) -> List[str]:
    parsed_context: List[str] = []
    if context and (isinstance(context, list) or isinstance(context, tuple)):
        for context_str in context:
            if not isinstance(context_str, str):
                pass
            else:
                parsed_context.append(context_str)
    return parsed_context


def post_test_run(traces: List[Trace], test_run_id: Optional[str]):
    # Accept single trace or list of traces
    if isinstance(traces, Trace):
        traces = [traces]

    api_test_cases: List[LLMApiTestCase] = []

    # Collect test cases from spans that have metric_collection
    for trace in traces:
        trace_api = trace_manager.create_trace_api(trace)

        def dfs(span: BaseSpan):
            if span.metric_collection:
                llm_test_case = LLMTestCase(
                    input=str(span.input),
                    actual_output=(
                        str(span.output) if span.output is not None else None
                    ),
                    expected_output=span.expected_output,
                    context=span.context,
                    retrieval_context=span.retrieval_context,
                    tools_called=span.tools_called,
                    expected_tools=span.expected_tools,
                )
                api_case = create_api_test_case(
                    test_case=llm_test_case,
                    trace=trace_api,
                    index=None,
                )
                if isinstance(api_case, LLMApiTestCase):
                    api_case.metric_collection = span.metric_collection
                    api_test_cases.append(api_case)

            for child in span.children or []:
                dfs(child)

        for root in trace.root_spans:
            dfs(root)

    # Prepare and post TestRun using the global test run manager
    test_run_manager = global_test_run_manager
    test_run_manager.create_test_run(identifier=test_run_id)
    test_run = test_run_manager.get_test_run()

    for case in api_test_cases:
        test_run.add_test_case(case)

    # return test_run_manager.post_test_run(test_run) TODO: add after test run with metric collection is implemented


def _normalize_pydantic_ai_messages(span: ReadableSpan) -> Optional[list]:
    try:
        raw = span.attributes.get("pydantic_ai.all_messages")
        if not raw:
            return None

        messages = raw
        if isinstance(messages, str):
            messages = json.loads(messages)
        elif isinstance(messages, tuple):
            messages = list(messages)

        if isinstance(messages, list):
            normalized = []
            for m in messages:
                if isinstance(m, str):
                    try:
                        m = json.loads(m)
                    except Exception:
                        pass
                normalized.append(m)
            return normalized
    except Exception:
        pass

    return None


def _extract_non_thinking_part_of_last_message(message: dict) -> dict:

    if isinstance(message, dict) and message.get("role") == "assistant":
        parts = message.get("parts")
        if parts:
            # Iterate from the last part
            for part in reversed(parts):
                if isinstance(part, dict) and part.get("type") == "text":
                    # Return a modified message with only the text content
                    return {"role": "assistant", "content": part.get("content")}
    return None


def check_pydantic_ai_agent_input_output(
    span: ReadableSpan,
) -> Tuple[Optional[Any], Optional[Any]]:
    input_val: list = []
    output_val: Optional[Any] = None

    # Get normalized messages once
    normalized = _normalize_pydantic_ai_messages(span)

    # Input (pydantic_ai.all_messages) - slice up to and including the first 'user' message
    if normalized:
        try:
            first_user_idx = None
            for i, m in enumerate(normalized):
                role = None
                if isinstance(m, dict):
                    role = m.get("role") or m.get("author")
                if role == "user":
                    first_user_idx = i
                    break

            input_val = (
                normalized
                if first_user_idx is None
                else normalized[: first_user_idx + 1]
            )
        except Exception:
            pass

    # Output (agent final_result)
    try:
        if span.attributes.get("confident.span.type") == "agent":
            output_val = span.attributes.get("final_result")
            if not output_val and normalized:
                output_val = _extract_non_thinking_part_of_last_message(
                    normalized[-1]
                )
    except Exception:
        pass

    system_instructions = []
    system_instruction_raw = span.attributes.get("gen_ai.system_instructions")
    if system_instruction_raw and isinstance(system_instruction_raw, str):
        system_instructions = _flatten_system_instructions(
            json.loads(system_instruction_raw)
        )

    input_val = _flatten_input(input_val)
    return system_instructions + input_val, output_val


def check_tool_output(span: ReadableSpan):
    try:
        return span.attributes.get("tool_response")
    except Exception:
        pass
    return None


def check_pydantic_ai_trace_input_output(
    span: ReadableSpan,
) -> Tuple[Optional[Any], Optional[Any]]:
    input_val: Optional[Any] = None
    output_val: Optional[Any] = None

    if not span.parent:
        input_val, output_val = check_pydantic_ai_agent_input_output(span)

    return input_val, output_val



import asyncio
from typing import Optional, Union, List

from deepeval.metrics import BaseConversationalMetric
from deepeval.models import DeepEvalBaseLLM
from deepeval.metrics.utils import (
    check_conversational_test_case_params,
    construct_verbose_logs,
    get_unit_interactions,
    trimAndLoadJson,
    initialize_model,
)
from deepeval.metrics.indicator import metric_progress_indicator
from deepeval.test_case import ConversationalTestCase, TurnParams
from deepeval.utils import get_or_create_event_loop, prettify_list
from deepeval.metrics.mcp.schema import Task, TaskScore
from deepeval.metrics.mcp.template import MCPTaskCompletionTemplate
from deepeval.errors import MissingTestCaseParamsError
from deepeval.metrics.api import metric_data_manager


class MCPTaskCompletionMetric(BaseConversationalMetric):
    _required_test_case_params = [
        TurnParams.ROLE,
        TurnParams.CONTENT,
    ]

    def __init__(
        self,
        threshold: float = 0.5,
        model: Optional[Union[str, DeepEvalBaseLLM]] = None,
        include_reason: bool = True,
        async_mode: bool = True,
        strict_mode: bool = False,
        verbose_mode: bool = False,
    ):
        self.threshold = 1 if strict_mode else threshold
        self.model, self.using_native_model = initialize_model(model)
        self.evaluation_model = self.model.get_model_name()
        self.include_reason = include_reason
        self.async_mode = async_mode
        self.strict_mode = strict_mode
        self.verbose_mode = verbose_mode

    def measure(
        self,
        test_case: ConversationalTestCase,
        _show_indicator: bool = True,
        _in_component: bool = False,
        _log_metric_to_confident: bool = True,
    ):
        check_conversational_test_case_params(
            test_case, self._required_test_case_params, self
        )

        self.evaluation_cost = 0 if self.using_native_model else None
        with metric_progress_indicator(
            self, _show_indicator=_show_indicator, _in_component=_in_component
        ):
            if self.async_mode:
                loop = get_or_create_event_loop()
                loop.run_until_complete(
                    self.a_measure(
                        test_case,
                        _show_indicator=False,
                        _in_component=_in_component,
                        _log_metric_to_confident=_log_metric_to_confident,
                    )
                )
            else:
                if not test_case.mcp_servers:
                    error_str = "'mcp_servers' in a conversational test case cannot be empty for the 'MCPTaskCompletionMetric' metric."
                    self.error = error_str
                    raise MissingTestCaseParamsError(error_str)

                self.unit_interactions = get_unit_interactions(test_case.turns)
                self.tasks = self._get_tasks(self.unit_interactions)
                self.task_scores = [
                    self._get_task_score(task) for task in self.tasks
                ]
                self.score = self._calculate_score(self.task_scores)
                self.reason = self._generate_reason(self.task_scores)
                self.scores_reasons_list = [
                    (task_score.score, task_score.reason)
                    for task_score in self.task_scores
                ]
                self.success = self.score >= self.threshold
                self.verbose_logs = construct_verbose_logs(
                    self,
                    steps=[
                        f"Tasks:\n{prettify_list(self.tasks)}",
                        f"Individual Scores & Reasons:\n{self.scores_reasons_list}",
                        f"Score: {self.score}",
                    ],
                )
                if _log_metric_to_confident:
                    metric_data_manager.post_metric_if_enabled(
                        self, test_case=test_case
                    )
            return self.score

    async def a_measure(
        self,
        test_case: ConversationalTestCase,
        _show_indicator: bool = True,
        _in_component: bool = False,
        _log_metric_to_confident: bool = True,
    ):
        check_conversational_test_case_params(
            test_case, self._required_test_case_params, self
        )

        self.evaluation_cost = 0 if self.using_native_model else None
        with metric_progress_indicator(
            self, async_mode=True, _show_indicator=_show_indicator
        ):
            if not test_case.mcp_servers:
                error_str = "'mcp_servers' in a conversational test case cannot be empty for the 'MCPTaskCompletionMetric' metric."
                self.error = error_str
                raise MissingTestCaseParamsError(error_str)

            self.unit_interactions = get_unit_interactions(test_case.turns)
            self.tasks = self._get_tasks(self.unit_interactions)
            self.task_scores = await asyncio.gather(
                *[self._a_get_task_score(task) for task in self.tasks]
            )
            self.scores_reasons_list = [
                (task_score.score, task_score.reason)
                for task_score in self.task_scores
            ]
            self.score = self._calculate_score(self.task_scores)
            self.reason = self._generate_reason(self.task_scores)
            self.success = self.score >= self.threshold
            self.verbose_logs = construct_verbose_logs(
                self,
                steps=[
                    f"Tasks:\n{prettify_list(self.tasks)}",
                    f"Individual Scores & Reasons:\n{prettify_list(self.scores_reasons_list)}",
                    f"Score: {self.score}",
                ],
            )
            if _log_metric_to_confident:
                metric_data_manager.post_metric_if_enabled(
                    self, test_case=test_case
                )

        return self.score

    def _generate_reason(self, task_scores: List[TaskScore]) -> str:
        reason = "["
        for task_score in task_scores:
            if task_score.score < self.threshold:
                reason += (
                    f"\nScore: {task_score.score}\n"
                    f"Reason: {task_score.reason}\n"
                )
        reason += "]"
        return reason

    def _get_task_score(self, task: Task) -> TaskScore:
        prompt = MCPTaskCompletionTemplate.get_task_completion_score(task)
        if self.using_native_model:
            res, cost = self.model.generate(prompt, schema=TaskScore)
            self.evaluation_cost += cost
            return res
        else:
            try:
                res: TaskScore = self.model.generate(prompt, schema=TaskScore)
                return res
            except TypeError:
                res = self.model.generate(prompt)
                data = trimAndLoadJson(res, self)
                return TaskScore(**data)

    async def _a_get_task_score(self, task: Task) -> TaskScore:
        prompt = MCPTaskCompletionTemplate.get_task_completion_score(task)
        if self.using_native_model:
            res, cost = await self.model.a_generate(prompt, schema=TaskScore)
            self.evaluation_cost += cost
            return res
        else:
            try:
                res: TaskScore = await self.model.a_generate(
                    prompt, schema=TaskScore
                )
                return res
            except TypeError:
                res = await self.model.a_generate(prompt)
                data = trimAndLoadJson(res, self)
                return TaskScore(**data)

    def _get_tasks(self, unit_interactions: List) -> List[Task]:
        tasks = []
        for unit_interaction in unit_interactions:
            if len(unit_interaction) <= 2:
                continue
            user_messages = ""
            for turn in unit_interaction:
                if turn.role == "user":
                    user_messages += turn.content + "\n"
                else:
                    break
            new_task = Task(task=user_messages, steps_taken=[])
            for turn in unit_interaction[1:]:
                if turn._mcp_interaction:
                    mcp_interaction = "Tools called by agent: \n"
                    if turn.mcp_tools_called is not None:
                        for tool in turn.mcp_tools_called:
                            mcp_interaction += (
                                f"\n<Tool Called>\n"
                                f"\n**This does not appear to user**\n"
                                f"Name: {tool.name}\n"
                                f"Args: {tool.args}\n"
                                f"Result: \n{tool.result.structuredContent['result']}\n"
                                f"</Tool Called>\n"
                            )
                    if turn.mcp_resources_called is not None:
                        for resource in turn.mcp_resources_called:
                            mcp_interaction += (
                                f"\n<Resource Called>\n"
                                f"\n**This does not appear to user**\n"
                                f"URI: {resource.uri}\n"
                                f"Result: {str(resource.result)}\n"
                                f"</Resource Called>\n"
                            )
                    if turn.mcp_prompts_called is not None:
                        for prompt in turn.mcp_prompts_called:
                            mcp_interaction += (
                                f"\n<Prompt Called>\n"
                                f"\n**This does not appear to user**\n"
                                f"Name: {prompt.name}\n"
                                f"Result: {str(prompt.result)}\n"
                                f"</Prompt Called>\n"
                            )
                    new_task.steps_taken.append(mcp_interaction)
                else:
                    new_task.steps_taken.append(
                        "Agent's response to user: \n" + turn.content
                    )
            tasks.append(new_task)
        return tasks

    def _calculate_score(self, scores: List[TaskScore]) -> float:
        total_score = sum(score.score for score in scores)
        return total_score / len(scores)

    def is_successful(self) -> bool:
        if self.error is not None:
            self.success = False
        else:
            try:
                self.score >= self.threshold
            except:
                self.success = False
        return self.success

    @property
    def __name__(self):
        return "MCP Task Completion"



from enum import Enum
from typing import Union

KEY_FILE: str = ".deepeval"
HIDDEN_DIR: str = ".deepeval"
PYTEST_RUN_TEST_NAME: str = "CONFIDENT_AI_RUN_TEST_NAME"
LOGIN_PROMPT = "\n✨👀 Looking for a place for your LLM test data to live 🏡❤️ ? Use [rgb(106,0,255)]Confident AI[/rgb(106,0,255)] to get & share testing reports, experiment with models/prompts, and catch regressions for your LLM system. Just run [cyan]'deepeval login'[/cyan] in the CLI."


CONFIDENT_TRACE_VERBOSE = "CONFIDENT_TRACE_VERBOSE"
CONFIDENT_TRACE_FLUSH = "CONFIDENT_TRACE_FLUSH"
CONFIDENT_TRACE_SAMPLE_RATE = "CONFIDENT_TRACE_SAMPLE_RATE"
CONFIDENT_TRACE_ENVIRONMENT = "CONFIDENT_TRACE_ENVIRONMENT"
CONFIDENT_TRACING_ENABLED = "CONFIDENT_TRACING_ENABLED"

CONFIDENT_METRIC_LOGGING_VERBOSE = "CONFIDENT_METRIC_LOGGING_VERBOSE"
CONFIDENT_METRIC_LOGGING_FLUSH = "CONFIDENT_METRIC_LOGGING_FLUSH"
CONFIDENT_METRIC_LOGGING_SAMPLE_RATE = "CONFIDENT_METRIC_LOGGING_SAMPLE_RATE"
CONFIDENT_METRIC_LOGGING_ENABLED = "CONFIDENT_METRIC_LOGGING_ENABLED"


CONFIDENT_OPEN_BROWSER = "CONFIDENT_OPEN_BROWSER"
CONFIDENT_TEST_CASE_BATCH_SIZE = "CONFIDENT_TEST_CASE_BATCH_SIZE"


class ProviderSlug(str, Enum):
    OPENAI = "openai"
    AZURE = "azure"
    ANTHROPIC = "anthropic"
    BEDROCK = "bedrock"
    DEEPSEEK = "deepseek"
    GOOGLE = "google"
    GROK = "grok"
    KIMI = "kimi"
    LITELLM = "litellm"
    LOCAL = "local"
    OLLAMA = "ollama"


def slugify(value: Union[str, ProviderSlug]) -> str:
    return (
        value.value
        if isinstance(value, ProviderSlug)
        else str(value).strip().lower()
    )


SUPPORTED_PROVIDER_SLUGS = frozenset(s.value for s in ProviderSlug)



"""
DeepEval CLI: Model Provider Configuration Commands

General behavior for all `set-*` / `unset-*` commands:

- Non-secret settings (model name, endpoint, deployment, toggles) are always
  persisted in the hidden `.deepeval/.deepeval` JSON store.
- Secrets (API keys) are **never** written to the JSON store.
- If `--save=dotenv[:path]` is passed, both secrets and non-secrets are
  written to the specified dotenv file (default: `.env.local`).
  Dotenv files should be git-ignored.
- If `--save` is not passed, only the JSON store is updated.
- When unsetting a provider, only that provider’s keys are removed.
  If another provider’s credentials remain (e.g. `OPENAI_API_KEY`), it
  may still be selected as the default.
"""

import os
from typing import Optional
from rich import print
from rich.markup import escape
import webbrowser
import threading
import random
import string
import socket
import typer
from enum import Enum
from pydantic import SecretStr
from deepeval.key_handler import (
    EmbeddingKeyValues,
    ModelKeyValues,
)
from deepeval.telemetry import capture_login_event, capture_view_event
from deepeval.cli.test import app as test_app
from deepeval.cli.server import start_server
from deepeval.config.settings import get_settings
from deepeval.utils import delete_file_if_exists, open_browser
from deepeval.test_run.test_run import (
    LATEST_TEST_RUN_FILE_PATH,
    global_test_run_manager,
)
from deepeval.cli.utils import (
    render_login_message,
    upload_and_open_link,
    PROD,
)
from deepeval.confident.api import (
    is_confident,
)

app = typer.Typer(name="deepeval")
app.add_typer(test_app, name="test")


class Regions(Enum):
    US = "US"
    EU = "EU"


def generate_pairing_code():
    """Generate a random pairing code."""
    return "".join(random.choices(string.ascii_uppercase + string.digits, k=6))


def find_available_port():
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("localhost", 0))  # Bind to port 0 to get an available port
        return s.getsockname()[1]


def is_openai_configured() -> bool:
    s = get_settings()
    v = s.OPENAI_API_KEY
    if isinstance(v, SecretStr):
        try:
            if v.get_secret_value().strip():
                return True
        except Exception:
            pass
    elif v and str(v).strip():
        return True
    env = os.getenv("OPENAI_API_KEY")
    return bool(env and env.strip())


@app.command(name="set-confident-region")
def set_confident_region_command(
    region: Regions = typer.Argument(
        ..., help="The data region to use (US or EU)"
    ),
    save: Optional[str] = typer.Option(
        None,
        "--save",
        help="Persist CLI parameters as environment variables in a dotenv file. "
        "Usage: --save=dotenv[:path] (default: .env.local)",
    ),
):
    """Set the Confident AI data region."""
    # Add flag emojis based on region
    flag = "🇺🇸" if region == Regions.US else "🇪🇺"

    settings = get_settings()
    with settings.edit(save=save) as edit_ctx:
        settings.CONFIDENT_REGION = region.value

    handled, path, _ = edit_ctx.result

    if not handled and save is not None:
        # invalid --save format (unsupported)
        print("Unsupported --save option. Use --save=dotenv[:path].")
    elif path:
        # persisted to a file
        print(
            f"Saved environment variables to {path} (ensure it's git-ignored)."
        )
    else:
        # updated in-memory & process env only
        print(
            "Settings updated for this session. To persist, use --save=dotenv[:path] "
            "(default .env.local) or set DEEPEVAL_DEFAULT_SAVE=dotenv:.env.local"
        )

    print(
        f":raising_hands: Congratulations! You're now using the {flag}  {region.value} data region for Confident AI."
    )


@app.command()
def login(
    api_key: str = typer.Option(
        "",
        help="API key to get from https://app.confident-ai.com. Required if you want to log events to the server.",
    ),
    confident_api_key: Optional[str] = typer.Option(
        None,
        "--confident-api-key",
        "-c",
        help="Confident API key (non-interactive). If omitted, you'll be prompted to enter one. In all cases the key is saved to a dotenv file (default: .env.local) unless overridden with --save.",
    ),
    save: Optional[str] = typer.Option(
        None,
        "--save",
        help="Where to persist settings. Format: dotenv[:path]. Defaults to .env.local. If omitted, login still writes to .env.local.",
    ),
):
    with capture_login_event() as span:
        completed = False
        try:
            # Resolve the key from CLI flag or interactive flow
            if confident_api_key:
                key = confident_api_key.strip()
            else:
                render_login_message()

                # Start the pairing server
                port = find_available_port()
                pairing_code = generate_pairing_code()
                pairing_thread = threading.Thread(
                    target=start_server,
                    args=(pairing_code, port, PROD),
                    daemon=True,
                )
                pairing_thread.start()

                # Open web url
                login_url = f"{PROD}/pair?code={pairing_code}&port={port}"
                webbrowser.open(login_url)
                print(
                    f"(open this link if your browser did not open: [link={PROD}]{PROD}[/link])"
                )

                # Manual fallback if still empty
                if api_key == "":
                    while True:
                        api_key = input("🔐 Enter your API Key: ").strip()
                        if api_key:
                            break
                        else:
                            print(
                                "❌ API Key cannot be empty. Please try again.\n"
                            )
                key = api_key.strip()

            settings = get_settings()
            save = save or settings.DEEPEVAL_DEFAULT_SAVE or "dotenv:.env.local"
            with settings.edit(save=save) as edit_ctx:
                settings.API_KEY = key
                settings.CONFIDENT_API_KEY = key

            handled, path, updated = edit_ctx.result

            if updated:
                if not handled and save is not None:
                    # invalid --save format (unsupported)
                    print(
                        "Unsupported --save option. Use --save=dotenv[:path]."
                    )
                elif path:
                    # persisted to a file
                    print(
                        f"Saved environment variables to {path} (ensure it's git-ignored)."
                    )

            completed = True
            print(
                "\n🎉🥳 Congratulations! You've successfully logged in! :raising_hands:"
            )
            print(
                "You're now using DeepEval with [rgb(106,0,255)]Confident AI[/rgb(106,0,255)]. "
                "Follow our quickstart tutorial here: "
                "[bold][link=https://www.confident-ai.com/docs/llm-evaluation/quickstart]"
                "https://www.confident-ai.com/docs/llm-evaluation/quickstart[/link][/bold]"
            )
        except Exception as e:
            completed = False
            print(f"Login failed: {e}")
        finally:
            if getattr(span, "set_attribute", None):
                span.set_attribute("completed", completed)


@app.command()
def logout(
    save: Optional[str] = typer.Option(
        None,
        "--save",
        help="Where to remove the saved key from. Use format dotenv[:path]. If omitted, uses DEEPEVAL_DEFAULT_SAVE or .env.local. The JSON keystore is always cleared.",
    )
):
    """
    Log out of Confident AI.

    Behavior:
    - Always clears the Confident API key from the JSON keystore and process env.
    - Also removes credentials from a dotenv file; defaults to DEEPEVAL_DEFAULT_SAVE if set, otherwise.env.local.
      Override the target with --save=dotenv[:path].
    """
    settings = get_settings()
    save = save or settings.DEEPEVAL_DEFAULT_SAVE or "dotenv:.env.local"
    with settings.edit(save=save) as edit_ctx:
        settings.API_KEY = None
        settings.CONFIDENT_API_KEY = None

    handled, path, updated = edit_ctx.result

    if updated:
        if not handled and save is not None:
            # invalid --save format (unsupported)
            print("Unsupported --save option. Use --save=dotenv[:path].")
        elif path:
            # persisted to a file
            print(f"Removed Confident AI key(s) from {path}.")

    delete_file_if_exists(LATEST_TEST_RUN_FILE_PATH)

    print("\n🎉🥳 You've successfully logged out! :raising_hands: ")


@app.command()
def view():
    with capture_view_event() as span:
        if is_confident():
            last_test_run_link = (
                global_test_run_manager.get_latest_test_run_link()
            )
            if last_test_run_link:
                print(f"🔗 View test run: {last_test_run_link}")
                open_browser(last_test_run_link)
            else:
                upload_and_open_link(_span=span)
        else:
            upload_and_open_link(_span=span)


@app.command(name="set-debug")
def set_debug(
    # Core verbosity
    log_level: Optional[str] = typer.Option(
        None,
        "--log-level",
        help="Global LOG_LEVEL (DEBUG|INFO|WARNING|ERROR|CRITICAL|NOTSET).",
    ),
    verbose: Optional[bool] = typer.Option(
        None, "--verbose/--no-verbose", help="Toggle DEEPEVAL_VERBOSE_MODE."
    ),
    # Retry logging dials
    retry_before_level: Optional[str] = typer.Option(
        None,
        "--retry-before-level",
        help="Log level before a retry attempt (DEBUG|INFO|WARNING|ERROR|CRITICAL|NOTSET or numeric).",
    ),
    retry_after_level: Optional[str] = typer.Option(
        None,
        "--retry-after-level",
        help="Log level after a retry attempt (DEBUG|INFO|WARNING|ERROR|CRITICAL|NOTSET or numeric).",
    ),
    # gRPC visibility
    grpc: Optional[bool] = typer.Option(
        None, "--grpc/--no-grpc", help="Toggle DEEPEVAL_GRPC_LOGGING."
    ),
    grpc_verbosity: Optional[str] = typer.Option(
        None,
        "--grpc-verbosity",
        help="Set GRPC_VERBOSITY (DEBUG|INFO|ERROR|NONE).",
    ),
    grpc_trace: Optional[str] = typer.Option(
        None,
        "--grpc-trace",
        help=(
            "Set GRPC_TRACE to comma-separated tracer names or glob patterns "
            "(e.g. 'tcp,http,secure_endpoint', '*' for all, 'list_tracers' to print available)."
        ),
    ),
    # Confident tracing
    trace_verbose: Optional[bool] = typer.Option(
        None,
        "--trace-verbose/--no-trace-verbose",
        help="Enable / disable CONFIDENT_TRACE_VERBOSE.",
    ),
    trace_env: Optional[str] = typer.Option(
        None,
        "--trace-env",
        help='Set CONFIDENT_TRACE_ENVIRONMENT ("development", "staging", "production", etc).',
    ),
    trace_flush: Optional[bool] = typer.Option(
        None,
        "--trace-flush/--no-trace-flush",
        help="Enable / disable  CONFIDENT_TRACE_FLUSH.",
    ),
    trace_sample_rate: Optional[float] = typer.Option(
        None,
        "--trace-sample-rate",
        help="Set CONFIDENT_TRACE_SAMPLE_RATE.",
    ),
    metric_logging_verbose: Optional[bool] = typer.Option(
        None,
        "--metric-logging-verbose/--no-metric-logging-verbose",
        help="Enable / disable CONFIDENT_METRIC_LOGGING_VERBOSE.",
    ),
    metric_logging_flush: Optional[bool] = typer.Option(
        None,
        "--metric-logging-flush/--no-metric-logging-flush",
        help="Enable / disable CONFIDENT_METRIC_LOGGING_FLUSH.",
    ),
    metric_logging_sample_rate: Optional[float] = typer.Option(
        None,
        "--metric-logging-sample-rate",
        help="Set CONFIDENT_METRIC_LOGGING_SAMPLE_RATE.",
    ),
    metric_logging_enabled: Optional[bool] = typer.Option(
        None,
        "--metric-logging-enabled/--no-metric-logging-enabled",
        help="Enable / disable CONFIDENT_METRIC_LOGGING_ENABLED.",
    ),
    # Advanced / potentially surprising
    error_reporting: Optional[bool] = typer.Option(
        None,
        "--error-reporting/--no-error-reporting",
        help="Enable / disable ERROR_REPORTING.",
    ),
    ignore_errors: Optional[bool] = typer.Option(
        None,
        "--ignore-errors/--no-ignore-errors",
        help="Enable / disable IGNORE_DEEPEVAL_ERRORS (not recommended in normal debugging).",
    ),
    # Persistence
    save: Optional[str] = typer.Option(
        None,
        "--save",
        help="Persist CLI parameters as environment variables in a dotenv file. "
        "Usage: --save=dotenv[:path] (default: .env.local)",
    ),
):
    """
    Configure verbose debug behavior for DeepEval.

    This command lets you mix-and-match verbosity flags (global LOG_LEVEL, verbose mode),
    retry logger levels, gRPC wire logging, and Confident trace toggles. Values apply
    immediately to the current process and can be persisted to a dotenv file with --save.

    Examples:
        deepeval set-debug --log-level DEBUG --verbose --grpc --retry-before-level DEBUG --retry-after-level INFO
        deepeval set-debug --trace-verbose --trace-env staging --save dotenv:.env.local
    """
    settings = get_settings()
    with settings.edit(save=save) as edit_ctx:
        # Core verbosity
        if log_level is not None:
            settings.LOG_LEVEL = log_level
        if verbose is not None:
            settings.DEEPEVAL_VERBOSE_MODE = verbose

        # Retry logging
        if retry_before_level is not None:
            settings.DEEPEVAL_RETRY_BEFORE_LOG_LEVEL = retry_before_level
        if retry_after_level is not None:
            settings.DEEPEVAL_RETRY_AFTER_LOG_LEVEL = retry_after_level

        # gRPC
        if grpc is not None:
            settings.DEEPEVAL_GRPC_LOGGING = grpc
        if grpc_verbosity is not None:
            settings.GRPC_VERBOSITY = grpc_verbosity
        if grpc_trace is not None:
            settings.GRPC_TRACE = grpc_trace

        # Confident tracing
        if trace_verbose is not None:
            settings.CONFIDENT_TRACE_VERBOSE = trace_verbose
        if trace_env is not None:
            settings.CONFIDENT_TRACE_ENVIRONMENT = trace_env
        if trace_flush is not None:
            settings.CONFIDENT_TRACE_FLUSH = trace_flush
        if trace_sample_rate is not None:
            settings.CONFIDENT_TRACE_SAMPLE_RATE = trace_sample_rate

        # Confident metrics
        if metric_logging_verbose is not None:
            settings.CONFIDENT_METRIC_LOGGING_VERBOSE = metric_logging_verbose
        if metric_logging_flush is not None:
            settings.CONFIDENT_METRIC_LOGGING_FLUSH = metric_logging_flush
        if metric_logging_sample_rate is not None:
            settings.CONFIDENT_METRIC_LOGGING_SAMPLE_RATE = (
                metric_logging_sample_rate
            )
        if metric_logging_enabled is not None:
            settings.CONFIDENT_METRIC_LOGGING_ENABLED = metric_logging_enabled

        # Advanced
        if error_reporting is not None:
            settings.ERROR_REPORTING = error_reporting
        if ignore_errors is not None:
            settings.IGNORE_DEEPEVAL_ERRORS = ignore_errors

    handled, path, updated = edit_ctx.result

    if not updated:
        # no changes were made, so there is nothing to do.
        return

    if not handled and save is not None:
        print("Unsupported --save option. Use --save=dotenv[:path].")
    elif path:
        print(
            f"Saved environment variables to {path} (ensure it's git-ignored)."
        )
    else:
        print(
            "Settings updated for this session. To persist, use --save=dotenv[:path] "
            "(default .env.local) or set DEEPEVAL_DEFAULT_SAVE=dotenv:.env.local"
        )

    print(":loud_sound: Debug options updated.")


@app.command(name="unset-debug")
def unset_debug(
    save: Optional[str] = typer.Option(
        None,
        "--save",
        help="Remove only the debug-related environment variables from a dotenv file. "
        "Usage: --save=dotenv[:path] (default: .env.local)",
    ),
):
    """
    Restore default behavior by unsetting debug related variables.

    Behavior:
    - Resets LOG_LEVEL back to 'info'.
    - Unsets DEEPEVAL_VERBOSE_MODE, retry log-level overrides, gRPC and Confident trace flags.
    - If --save is provided (or DEEPEVAL_DEFAULT_SAVE is set), removes these keys from the target dotenv file.
    """
    settings = get_settings()
    with settings.edit(save=save) as edit_ctx:
        # Back to normal global level
        settings.LOG_LEVEL = "info"
        settings.CONFIDENT_TRACE_ENVIRONMENT = "development"
        settings.CONFIDENT_TRACE_VERBOSE = True
        settings.CONFIDENT_METRIC_LOGGING_VERBOSE = True
        settings.CONFIDENT_METRIC_LOGGING_ENABLED = True

        # Clear optional toggles/overrides
        settings.DEEPEVAL_VERBOSE_MODE = None
        settings.DEEPEVAL_RETRY_BEFORE_LOG_LEVEL = None
        settings.DEEPEVAL_RETRY_AFTER_LOG_LEVEL = None

        settings.DEEPEVAL_GRPC_LOGGING = None
        settings.GRPC_VERBOSITY = None
        settings.GRPC_TRACE = None

        settings.CONFIDENT_TRACE_FLUSH = None
        settings.CONFIDENT_METRIC_LOGGING_FLUSH = None

        settings.ERROR_REPORTING = None
        settings.IGNORE_DEEPEVAL_ERRORS = None

    handled, path, _ = edit_ctx.result

    if not handled and save is not None:
        print("Unsupported --save option. Use --save=dotenv[:path].")
    elif path:
        print(f"Removed debug-related environment variables from {path}.")
    else:
        print("Debug settings reverted to defaults for this session.")

    print(":mute: Debug options unset.")


#############################################
# OpenAI Integration ########################
#############################################


@app.command(name="set-openai")
def set_openai_env(
    model: str = typer.Option(..., "--model", help="OpenAI model name"),
    cost_per_input_token: Optional[float] = typer.Option(
        None,
        "--cost_per_input_token",
        help=(
            "USD per input token. Optional for known OpenAI models (pricing is preconfigured); "
            "REQUIRED if you use a custom/unsupported model."
        ),
    ),
    cost_per_output_token: Optional[float] = typer.Option(
        None,
        "--cost_per_output_token",
        help=(
            "USD per output token. Optional for known OpenAI models (pricing is preconfigured); "
            "REQUIRED if you use a custom/unsupported model."
        ),
    ),
    save: Optional[str] = typer.Option(
        None,
        "--save",
        help="Persist CLI parameters as environment variables in a dotenv file. "
        "Usage: --save=dotenv[:path] (default: .env.local)",
    ),
):
    """
    Configure OpenAI as the active LLM provider.

    What this does:
    - Sets the active provider flag to `USE_OPENAI_MODEL`.
    - Persists the selected model name and any cost overrides in the JSON store.
    - secrets are never written to `.deepeval/.deepeval` (JSON).

    Pricing rules:
    - If `model` is a known OpenAI model, you may omit costs (built‑in pricing is used).
    - If `model` is custom/unsupported, you must provide both
      `--cost_per_input_token` and `--cost_per_output_token`.

    Secrets & saving:
    - Set your `OPENAI_API_KEY` via environment or a dotenv file.
    - Pass `--save=dotenv[:path]` to write configuration to a dotenv file
      (default: `.env.local`). This command does not set or persist OPENAI_API_KEY. Set it
      via your environment or a dotenv file (e.g., add OPENAI_API_KEY=... to .env.local)
      before running this command, or manage it with whatever command you use for secrets.

    Args:
        model: OpenAI model name, such as `gpt-4o-mini`.
        cost_per_input_token: USD per input token (optional for known models).
        cost_per_output_token: USD per output token (optional for known models).
        save: Persist config (and supported secrets) to a dotenv file; format `dotenv[:path]`.

    Example:
        deepeval set-openai \\
          --model gpt-4o-mini \\
          --cost_per_input_token 0.0005 \\
          --cost_per_output_token 0.0015 \\
          --save dotenv:.env.local
    """
    settings = get_settings()
    with settings.edit(save=save) as edit_ctx:
        edit_ctx.switch_model_provider(ModelKeyValues.USE_OPENAI_MODEL)
        settings.OPENAI_MODEL_NAME = model
        if cost_per_input_token is not None:
            settings.OPENAI_COST_PER_INPUT_TOKEN = cost_per_input_token
        if cost_per_output_token is not None:
            settings.OPENAI_COST_PER_OUTPUT_TOKEN = cost_per_output_token

    handled, path, _ = edit_ctx.result

    if not handled and save is not None:
        # invalid --save format (unsupported)
        print("Unsupported --save option. Use --save=dotenv[:path].")
    elif path:
        # persisted to a file
        print(
            f"Saved environment variables to {path} (ensure it's git-ignored)."
        )
    else:
        # updated in-memory & process env only
        print(
            "Tip: persist these settings to a dotenv file with --save=dotenv[:path] (default .env.local) "
            "or set DEEPEVAL_DEFAULT_SAVE=dotenv:.env.local"
        )

    print(
        f":raising_hands: Congratulations! You're now using OpenAI's `{escape(model)}` for all evals that require an LLM."
    )


@app.command(name="unset-openai")
def unset_openai_env(
    save: Optional[str] = typer.Option(
        None,
        "--save",
        help="Remove only the OpenAI related environment variables from a dotenv file. "
        "Usage: --save=dotenv[:path] (default: .env.local)",
    ),
):
    """
    Unset OpenAI as the active provider.

    Behavior:
    - Removes OpenAI keys (model, costs, toggle) from the JSON store.
    - If `--save` is provided, removes those keys from the specified dotenv file.
    - After unsetting, if `OPENAI_API_KEY` is still set in the environment,
      OpenAI may still be usable by default. Otherwise, no active provider is configured.

    Args:
        --save: Remove OpenAI keys from the given dotenv file as well.

    Example:
        deepeval unset-openai --save dotenv:.env.local
    """

    settings = get_settings()
    with settings.edit(save=save) as edit_ctx:
        settings.OPENAI_MODEL_NAME = None
        settings.OPENAI_COST_PER_INPUT_TOKEN = None
        settings.OPENAI_COST_PER_OUTPUT_TOKEN = None
        settings.USE_OPENAI_MODEL = None

    handled, path, _ = edit_ctx.result

    if not handled and save is not None:
        # invalid --save format (unsupported)
        print("Unsupported --save option. Use --save=dotenv[:path].")
    elif path:
        # persisted to a file
        print(f"Removed OpenAI environment variables from {path}.")

    if is_openai_configured():
        print(
            ":raised_hands: OpenAI will still be used by default because OPENAI_API_KEY is set."
        )
    else:
        print(
            "OpenAI has been unset. No active provider is configured. Set one with the CLI, or add credentials to .env[.local]."
        )


#############################################
# Azure Integration ########################
#############################################


@app.command(name="set-azure-openai")
def set_azure_openai_env(
    azure_openai_api_key: str = typer.Option(
        ...,
        "--openai-api-key",
        help="Azure OpenAI API key",
    ),
    azure_openai_endpoint: str = typer.Option(
        ..., "--openai-endpoint", help="Azure OpenAI endpoint"
    ),
    openai_api_version: str = typer.Option(
        ..., "--openai-api-version", help="OpenAI API version"
    ),
    openai_model_name: str = typer.Option(
        ..., "--openai-model-name", help="OpenAI model name"
    ),
    azure_deployment_name: str = typer.Option(
        ..., "--deployment-name", help="Azure deployment name"
    ),
    azure_model_version: Optional[str] = typer.Option(
        None, "--model-version", help="Azure model version (optional)"
    ),
    save: Optional[str] = typer.Option(
        None,
        "--save",
        help="Persist CLI parameters as environment variables in a dotenv file. "
        "Usage: --save=dotenv[:path] (default: .env.local)",
    ),
):
    settings = get_settings()
    with settings.edit(save=save) as edit_ctx:
        edit_ctx.switch_model_provider(ModelKeyValues.USE_AZURE_OPENAI)
        settings.AZURE_OPENAI_API_KEY = azure_openai_api_key
        settings.AZURE_OPENAI_ENDPOINT = azure_openai_endpoint
        settings.OPENAI_API_VERSION = openai_api_version
        settings.AZURE_DEPLOYMENT_NAME = azure_deployment_name
        settings.AZURE_MODEL_NAME = openai_model_name
        if azure_model_version is not None:
            settings.AZURE_MODEL_VERSION = azure_model_version

    handled, path, _ = edit_ctx.result

    if not handled and save is not None:
        # invalid --save format (unsupported)
        print("Unsupported --save option. Use --save=dotenv[:path].")
    elif path:
        # persisted to a file
        print(
            f"Saved environment variables to {path} (ensure it's git-ignored)."
        )
    else:
        # updated in-memory & process env only
        print(
            "Settings updated for this session. To persist, use --save=dotenv[:path] "
            "(default .env.local) or set DEEPEVAL_DEFAULT_SAVE=dotenv:.env.local"
        )

    print(
        f":raising_hands: Congratulations! You're now using Azure OpenAI's `{escape(openai_model_name)}` for all evals that require an LLM."
    )


@app.command(name="set-azure-openai-embedding")
def set_azure_openai_embedding_env(
    azure_embedding_deployment_name: str = typer.Option(
        ...,
        "--embedding-deployment-name",
        help="Azure embedding deployment name",
    ),
    save: Optional[str] = typer.Option(
        None,
        "--save",
        help="Persist CLI parameters as environment variables in a dotenv file. "
        "Usage: --save=dotenv[:path] (default: .env.local)",
    ),
):
    settings = get_settings()
    with settings.edit(save=save) as edit_ctx:
        edit_ctx.switch_model_provider(
            EmbeddingKeyValues.USE_AZURE_OPENAI_EMBEDDING
        )
        settings.AZURE_EMBEDDING_DEPLOYMENT_NAME = (
            azure_embedding_deployment_name
        )

    handled, path, _ = edit_ctx.result

    if not handled and save is not None:
        # invalid --save format (unsupported)
        print("Unsupported --save option. Use --save=dotenv[:path].")
    elif path:
        # persisted to a file
        print(
            f"Saved environment variables to {path} (ensure it's git-ignored)."
        )
    else:
        # updated in-memory & process env only
        print(
            "Settings updated for this session. To persist, use --save=dotenv[:path] "
            "(default .env.local) or set DEEPEVAL_DEFAULT_SAVE=dotenv:.env.local"
        )

    print(
        ":raising_hands: Congratulations! You're now using Azure OpenAI Embeddings within DeepEval."
    )


@app.command(name="unset-azure-openai")
def unset_azure_openai_env(
    save: Optional[str] = typer.Option(
        None,
        "--save",
        help="Remove only the Azure OpenAI–related environment variables from a dotenv file. "
        "Usage: --save=dotenv[:path] (default: .env.local)",
    )
):
    settings = get_settings()
    with settings.edit(save=save) as edit_ctx:
        settings.AZURE_OPENAI_API_KEY = None
        settings.AZURE_OPENAI_ENDPOINT = None
        settings.OPENAI_API_VERSION = None
        settings.AZURE_DEPLOYMENT_NAME = None
        settings.AZURE_MODEL_NAME = None
        settings.AZURE_MODEL_VERSION = None
        settings.USE_AZURE_OPENAI = None

    handled, path, _ = edit_ctx.result

    if not handled and save is not None:
        # invalid --save format (unsupported)
        print("Unsupported --save option. Use --save=dotenv[:path].")
    elif path:
        # persisted to a file
        print(f"Removed Azure OpenAI environment variables from {path}.")

    if is_openai_configured():
        print(
            ":raised_hands: OpenAI will still be used by default because OPENAI_API_KEY is set."
        )
    else:
        print(
            "Azure OpenAI has been unset. No active provider is configured. Set one with the CLI, or add credentials to .env[.local]."
        )


@app.command(name="unset-azure-openai-embedding")
def unset_azure_openai_embedding_env(
    save: Optional[str] = typer.Option(
        None,
        "--save",
        help="Remove only the Azure OpenAI embedding related environment variables from a dotenv file. "
        "Usage: --save=dotenv[:path] (default: .env.local)",
    ),
):
    settings = get_settings()
    with settings.edit(save=save) as edit_ctx:
        settings.AZURE_EMBEDDING_DEPLOYMENT_NAME = None
        settings.USE_AZURE_OPENAI_EMBEDDING = None

    handled, path, _ = edit_ctx.result

    if not handled and save is not None:
        # invalid --save format (unsupported)
        print("Unsupported --save option. Use --save=dotenv[:path].")
    elif path:
        # persisted to a file
        print(
            f"Removed Azure OpenAI embedding environment variables from {path}."
        )

    if is_openai_configured():
        print(
            ":raised_hands: Regular OpenAI embeddings will still be used by default because OPENAI_API_KEY is set."
        )
    else:
        print(
            "The Azure OpenAI embedding model configuration has been removed. No model is currently configured, but you can set one with the CLI or add credentials to .env[.local]."
        )


#############################################
# Ollama Integration ########################
#############################################


@app.command(name="set-ollama")
def set_ollama_model_env(
    model_name: str = typer.Argument(..., help="Name of the Ollama model"),
    base_url: str = typer.Option(
        "http://localhost:11434",
        "-b",
        "--base-url",
        help="Base URL for the local model API",
    ),
    save: Optional[str] = typer.Option(
        None,
        "--save",
        help="Persist CLI parameters as environment variables in a dotenv file. "
        "Usage: --save=dotenv[:path] (default: .env.local)",
    ),
):
    settings = get_settings()
    with settings.edit(save=save) as edit_ctx:
        edit_ctx.switch_model_provider(ModelKeyValues.USE_LOCAL_MODEL)
        settings.LOCAL_MODEL_API_KEY = "ollama"
        settings.LOCAL_MODEL_NAME = model_name
        settings.LOCAL_MODEL_BASE_URL = base_url

    handled, path, _ = edit_ctx.result

    if not handled and save is not None:
        # invalid --save format (unsupported)
        print("Unsupported --save option. Use --save=dotenv[:path].")
    elif path:
        # persisted to a file
        print(
            f"Saved environment variables to {path} (ensure it's git-ignored)."
        )
    else:
        # updated in-memory & process env only
        print(
            "Settings updated for this session. To persist, use --save=dotenv[:path] "
            "(default .env.local) or set DEEPEVAL_DEFAULT_SAVE=dotenv:.env.local"
        )

    print(
        f":raising_hands: Congratulations! You're now using a local Ollama model `{escape(model_name)}` for all evals that require an LLM."
    )


@app.command(name="unset-ollama")
def unset_ollama_model_env(
    save: Optional[str] = typer.Option(
        None,
        "--save",
        help="Remove only the Ollama related environment variables from a dotenv file. "
        "Usage: --save=dotenv[:path] (default: .env.local)",
    ),
):
    settings = get_settings()
    with settings.edit(save=save) as edit_ctx:
        settings.LOCAL_MODEL_API_KEY = None
        settings.LOCAL_MODEL_NAME = None
        settings.LOCAL_MODEL_BASE_URL = None
        settings.USE_LOCAL_MODEL = None

    handled, path, _ = edit_ctx.result

    if not handled and save is not None:
        # invalid --save format (unsupported)
        print("Unsupported --save option. Use --save=dotenv[:path].")
    elif path:
        # persisted to a file
        print(f"Removed local Ollama environment variables from {path}.")

    if is_openai_configured():
        print(
            ":raised_hands: OpenAI will still be used by default because OPENAI_API_KEY is set."
        )
    else:
        print(
            "The local Ollama model configuration has been removed. No model is currently configured, but you can set one with the CLI or add credentials to .env[.local]."
        )


@app.command(name="set-ollama-embeddings")
def set_ollama_embeddings_env(
    model_name: str = typer.Argument(
        ..., help="Name of the Ollama embedding model"
    ),
    base_url: str = typer.Option(
        "http://localhost:11434",
        "-b",
        "--base-url",
        help="Base URL for the Ollama embedding model API",
    ),
    save: Optional[str] = typer.Option(
        None,
        "--save",
        help="Persist CLI parameters as environment variables in a dotenv file. "
        "Usage: --save=dotenv[:path] (default: .env.local)",
    ),
):
    settings = get_settings()
    with settings.edit(save=save) as edit_ctx:
        edit_ctx.switch_model_provider(EmbeddingKeyValues.USE_LOCAL_EMBEDDINGS)
        settings.LOCAL_EMBEDDING_API_KEY = "ollama"
        settings.LOCAL_EMBEDDING_MODEL_NAME = model_name
        settings.LOCAL_EMBEDDING_BASE_URL = base_url

    handled, path, _ = edit_ctx.result

    if not handled and save is not None:
        # invalid --save format (unsupported)
        print("Unsupported --save option. Use --save=dotenv[:path].")
    elif path:
        # persisted to a file
        print(
            f"Saved environment variables to {path} (ensure it's git-ignored)."
        )
    else:
        # updated in-memory & process env only
        print(
            "Settings updated for this session. To persist, use --save=dotenv[:path] "
            "(default .env.local) or set DEEPEVAL_DEFAULT_SAVE=dotenv:.env.local"
        )

    print(
        f":raising_hands: Congratulations! You're now using the Ollama embedding model `{escape(model_name)}` for all evals that require text embeddings."
    )


@app.command(name="unset-ollama-embeddings")
def unset_ollama_embeddings_env(
    save: Optional[str] = typer.Option(
        None,
        "--save",
        help="Remove only the Ollama embedding related environment variables from a dotenv file. "
        "Usage: --save=dotenv[:path] (default: .env.local)",
    ),
):

    settings = get_settings()
    with settings.edit(save=save) as edit_ctx:
        settings.LOCAL_EMBEDDING_API_KEY = None
        settings.LOCAL_EMBEDDING_MODEL_NAME = None
        settings.LOCAL_EMBEDDING_BASE_URL = None
        settings.USE_LOCAL_EMBEDDINGS = None

    handled, path, _ = edit_ctx.result

    if not handled and save is not None:
        # invalid --save format (unsupported)
        print("Unsupported --save option. Use --save=dotenv[:path].")
    elif path:
        # persisted to a file
        print(
            f"Removed local Ollama embedding environment variables from {path}."
        )

    if is_openai_configured():
        print(
            ":raised_hands: Regular OpenAI embeddings will still be used by default because OPENAI_API_KEY is set."
        )
    else:
        print(
            "The local Ollama embedding model configuration has been removed. No model is currently configured, but you can set one with the CLI or add credentials to .env[.local]."
        )


#############################################
# Local Model Integration ###################
#############################################


@app.command(name="set-local-model")
def set_local_model_env(
    model_name: str = typer.Option(
        ..., "--model-name", help="Name of the local model"
    ),
    base_url: str = typer.Option(
        ..., "--base-url", help="Base URL for the local model API"
    ),
    api_key: Optional[str] = typer.Option(
        None,
        "--api-key",
        help="API key for the local model. Persisted to dotenv if --save is used; never written to the legacy JSON keystore.",
    ),
    model_format: Optional[str] = typer.Option(
        "json",
        "--format",
        help="Format of the response from the local model (default: json)",
    ),
    save: Optional[str] = typer.Option(
        None,
        "--save",
        help="Persist CLI parameters as environment variables in a dotenv file. "
        "Usage: --save=dotenv[:path] (default: .env.local)",
    ),
):
    settings = get_settings()
    with settings.edit(save=save) as edit_ctx:
        edit_ctx.switch_model_provider(ModelKeyValues.USE_LOCAL_MODEL)
        settings.LOCAL_MODEL_NAME = model_name
        settings.LOCAL_MODEL_BASE_URL = base_url
        if model_format:
            settings.LOCAL_MODEL_FORMAT = model_format
        if api_key:
            settings.LOCAL_MODEL_API_KEY = api_key

    handled, path, _ = edit_ctx.result

    if not handled and save is not None:
        # invalid --save format (unsupported)
        print("Unsupported --save option. Use --save=dotenv[:path].")
    elif path:
        # persisted to a file
        print(
            f"Saved environment variables to {path} (ensure it's git-ignored)."
        )
    else:
        # updated in-memory & process env only
        print(
            "Settings updated for this session. To persist, use --save=dotenv[:path] "
            "(default .env.local) or set DEEPEVAL_DEFAULT_SAVE=dotenv:.env.local"
        )

    print(
        f":raising_hands: Congratulations! You're now using a local model `{escape(model_name)}` for all evals that require an LLM."
    )


@app.command(name="unset-local-model")
def unset_local_model_env(
    save: Optional[str] = typer.Option(
        None,
        "--save",
        help="Remove only the local model related environment variables from a dotenv file. "
        "Usage: --save=dotenv[:path] (default: .env.local)",
    ),
):
    settings = get_settings()
    with settings.edit(save=save) as edit_ctx:
        settings.LOCAL_MODEL_API_KEY = None
        settings.LOCAL_MODEL_NAME = None
        settings.LOCAL_MODEL_BASE_URL = None
        settings.LOCAL_MODEL_FORMAT = None
        settings.USE_LOCAL_MODEL = None

    handled, path, _ = edit_ctx.result

    if not handled and save is not None:
        # invalid --save format (unsupported)
        print("Unsupported --save option. Use --save=dotenv[:path].")
    elif path:
        # persisted to a file
        print(f"Removed local model environment variables from {path}.")

    if is_openai_configured():
        print(
            ":raised_hands: OpenAI will still be used by default because OPENAI_API_KEY is set."
        )
    else:
        print(
            "The local model configuration has been removed. No model is currently configured, but you can set one with the CLI or add credentials to .env[.local]."
        )


#############################################
# Grok Model Integration ####################
#############################################


@app.command(name="set-grok")
def set_grok_model_env(
    model_name: str = typer.Option(
        ..., "--model-name", help="Name of the Grok model"
    ),
    api_key: str = typer.Option(
        ...,
        "--api-key",
        help="API key for the Grok model. Persisted to dotenv if --save is used; never written to the legacy JSON keystore.",
    ),
    temperature: float = typer.Option(
        0, "--temperature", help="Temperature for the Grok model"
    ),
    save: Optional[str] = typer.Option(
        None,
        "--save",
        help="Persist CLI parameters as environment variables in a dotenv file. "
        "Usage: --save=dotenv[:path] (default: .env.local)",
    ),
):
    settings = get_settings()
    with settings.edit(save=save) as edit_ctx:
        edit_ctx.switch_model_provider(ModelKeyValues.USE_GROK_MODEL)
        settings.GROK_API_KEY = api_key
        settings.GROK_MODEL_NAME = model_name
        settings.TEMPERATURE = temperature

    handled, path, _ = edit_ctx.result

    if not handled and save is not None:
        # invalid --save format (unsupported)
        print("Unsupported --save option. Use --save=dotenv[:path].")
    elif path:
        # persisted to a file
        print(
            f"Saved environment variables to {path} (ensure it's git-ignored)."
        )
    else:
        # updated in-memory & process env only
        print(
            "Settings updated for this session. To persist, use --save=dotenv[:path] "
            "(default .env.local) or set DEEPEVAL_DEFAULT_SAVE=dotenv:.env.local"
        )

    print(
        f":raising_hands: Congratulations! You're now using a Grok's `{escape(model_name)}` for all evals that require an LLM."
    )


@app.command(name="unset-grok")
def unset_grok_model_env(
    save: Optional[str] = typer.Option(
        None,
        "--save",
        help="Remove only the Grok model related environment variables from a dotenv file. "
        "Usage: --save=dotenv[:path] (default: .env.local)",
    ),
):
    settings = get_settings()
    with settings.edit(save=save) as edit_ctx:
        settings.GROK_API_KEY = None
        settings.GROK_MODEL_NAME = None
        settings.TEMPERATURE = None
        settings.USE_GROK_MODEL = None

    handled, path, _ = edit_ctx.result

    if not handled and save is not None:
        # invalid --save format (unsupported)
        print("Unsupported --save option. Use --save=dotenv[:path].")
    elif path:
        # persisted to a file
        print(f"Removed Grok model environment variables from {path}.")

    if is_openai_configured():
        print(
            ":raised_hands: OpenAI will still be used by default because OPENAI_API_KEY is set."
        )
    else:
        print(
            "The Grok model configuration has been removed. No model is currently configured, but you can set one with the CLI or add credentials to .env[.local]."
        )


#############################################
# Moonshot Model Integration ################
#############################################


@app.command(name="set-moonshot")
def set_moonshot_model_env(
    model_name: str = typer.Option(
        ..., "--model-name", help="Name of the Moonshot model"
    ),
    api_key: str = typer.Option(
        ...,
        "--api-key",
        help="API key for the Moonshot model. Persisted to dotenv if --save is used; never written to the legacy JSON keystore.",
    ),
    temperature: float = typer.Option(
        0, "--temperature", help="Temperature for the Moonshot model"
    ),
    save: Optional[str] = typer.Option(
        None,
        "--save",
        help="Persist CLI parameters as environment variables in a dotenv file. "
        "Usage: --save=dotenv[:path] (default: .env.local)",
    ),
):
    settings = get_settings()
    with settings.edit(save=save) as edit_ctx:
        edit_ctx.switch_model_provider(ModelKeyValues.USE_MOONSHOT_MODEL)
        settings.MOONSHOT_API_KEY = api_key
        settings.MOONSHOT_MODEL_NAME = model_name
        settings.TEMPERATURE = temperature

    handled, path, _ = edit_ctx.result

    if not handled and save is not None:
        # invalid --save format (unsupported)
        print("Unsupported --save option. Use --save=dotenv[:path].")
    elif path:
        # persisted to a file
        print(
            f"Saved environment variables to {path} (ensure it's git-ignored)."
        )
    else:
        # updated in-memory & process env only
        print(
            "Settings updated for this session. To persist, use --save=dotenv[:path] "
            "(default .env.local) or set DEEPEVAL_DEFAULT_SAVE=dotenv:.env.local"
        )

    print(
        f":raising_hands: Congratulations! You're now using Moonshot's `{escape(model_name)}` for all evals that require an LLM."
    )


@app.command(name="unset-moonshot")
def unset_moonshot_model_env(
    save: Optional[str] = typer.Option(
        None,
        "--save",
        help="Remove only the Moonshot model related environment variables from a dotenv file. "
        "Usage: --save=dotenv[:path] (default: .env.local)",
    ),
):
    settings = get_settings()
    with settings.edit(save=save) as edit_ctx:
        settings.MOONSHOT_API_KEY = None
        settings.MOONSHOT_MODEL_NAME = None
        settings.TEMPERATURE = None
        settings.USE_MOONSHOT_MODEL = None

    handled, path, _ = edit_ctx.result

    if not handled and save is not None:
        # invalid --save format (unsupported)
        print("Unsupported --save option. Use --save=dotenv[:path].")
    elif path:
        # persisted to a file
        print(f"Removed Moonshot model environment variables from {path}.")

    if is_openai_configured():
        print(
            ":raised_hands: OpenAI will still be used by default because OPENAI_API_KEY is set."
        )
    else:
        print(
            "The Moonshot model configuration has been removed. No model is currently configured, but you can set one with the CLI or add credentials to .env[.local]."
        )


#############################################
# DeepSeek Model Integration ################
#############################################


@app.command(name="set-deepseek")
def set_deepseek_model_env(
    model_name: str = typer.Option(
        ..., "--model-name", help="Name of the DeepSeek model"
    ),
    api_key: str = typer.Option(
        ...,
        "--api-key",
        help="API key for the DeepSeek model. Persisted to dotenv if --save is used; never written to the legacy JSON keystore.",
    ),
    temperature: float = typer.Option(
        0, "--temperature", help="Temperature for the DeepSeek model"
    ),
    save: Optional[str] = typer.Option(
        None,
        "--save",
        help="Persist CLI parameters as environment variables in a dotenv file. "
        "Usage: --save=dotenv[:path] (default: .env.local)",
    ),
):
    settings = get_settings()
    with settings.edit(save=save) as edit_ctx:
        edit_ctx.switch_model_provider(ModelKeyValues.USE_DEEPSEEK_MODEL)
        settings.DEEPSEEK_API_KEY = api_key
        settings.DEEPSEEK_MODEL_NAME = model_name
        settings.TEMPERATURE = temperature

    handled, path, _ = edit_ctx.result

    if not handled and save is not None:
        # invalid --save format (unsupported)
        print("Unsupported --save option. Use --save=dotenv[:path].")
    elif path:
        # persisted to a file
        print(
            f"Saved environment variables to {path} (ensure it's git-ignored)."
        )
    else:
        # updated in-memory & process env only
        print(
            "Settings updated for this session. To persist, use --save=dotenv[:path] "
            "(default .env.local) or set DEEPEVAL_DEFAULT_SAVE=dotenv:.env.local"
        )

    print(
        f":raising_hands: Congratulations! You're now using DeepSeek's `{escape(model_name)}` for all evals that require an LLM."
    )


@app.command(name="unset-deepseek")
def unset_deepseek_model_env(
    save: Optional[str] = typer.Option(
        None,
        "--save",
        help="Remove only the DeepSeek model related environment variables from a dotenv file. "
        "Usage: --save=dotenv[:path] (default: .env.local)",
    ),
):
    settings = get_settings()
    with settings.edit(save=save) as edit_ctx:
        settings.DEEPSEEK_API_KEY = None
        settings.DEEPSEEK_MODEL_NAME = None
        settings.TEMPERATURE = None
        settings.USE_DEEPSEEK_MODEL = None

    handled, path, _ = edit_ctx.result

    if not handled and save is not None:
        # invalid --save format (unsupported)
        print("Unsupported --save option. Use --save=dotenv[:path].")
    elif path:
        # persisted to a file
        print(f"Removed DeepSeek model environment variables from {path}.")

    if is_openai_configured():
        print(
            ":raised_hands: OpenAI will still be used by default because OPENAI_API_KEY is set."
        )
    else:
        print(
            "The DeepSeek model configuration has been removed. No model is currently configured, but you can set one with the CLI or add credentials to .env[.local]."
        )


#############################################
# Local Embedding Model Integration #########
#############################################


@app.command(name="set-local-embeddings")
def set_local_embeddings_env(
    model_name: str = typer.Option(
        ..., "--model-name", help="Name of the local embeddings model"
    ),
    base_url: str = typer.Option(
        ..., "--base-url", help="Base URL for the local embeddings API"
    ),
    api_key: Optional[str] = typer.Option(
        None,
        "--api-key",
        help="API key for the local embeddings. Persisted to dotenv if --save is used; never written to the legacy JSON keystore.",
    ),
    save: Optional[str] = typer.Option(
        None,
        "--save",
        help="Persist CLI parameters as environment variables in a dotenv file. "
        "Usage: --save=dotenv[:path] (default: .env.local)",
    ),
):
    settings = get_settings()
    with settings.edit(save=save) as edit_ctx:
        edit_ctx.switch_model_provider(EmbeddingKeyValues.USE_LOCAL_EMBEDDINGS)
        settings.LOCAL_EMBEDDING_MODEL_NAME = model_name
        settings.LOCAL_EMBEDDING_BASE_URL = base_url
        if api_key:
            settings.LOCAL_EMBEDDING_API_KEY = api_key

    handled, path, _ = edit_ctx.result

    if not handled and save is not None:
        # invalid --save format (unsupported)
        print("Unsupported --save option. Use --save=dotenv[:path].")
    elif path:
        # persisted to a file
        print(
            f"Saved environment variables to {path} (ensure it's git-ignored)."
        )
    else:
        # updated in-memory & process env only
        print(
            "Settings updated for this session. To persist, use --save=dotenv[:path] "
            "(default .env.local) or set DEEPEVAL_DEFAULT_SAVE=dotenv:.env.local"
        )

    print(
        f":raising_hands: Congratulations! You're now using the local embedding model `{escape(model_name)}` for all evals that require text embeddings."
    )


@app.command(name="unset-local-embeddings")
def unset_local_embeddings_env(
    save: Optional[str] = typer.Option(
        None,
        "--save",
        help="Remove only the local embedding related environment variables from a dotenv file. "
        "Usage: --save=dotenv[:path] (default: .env.local)",
    ),
):
    settings = get_settings()
    with settings.edit(save=save) as edit_ctx:
        settings.LOCAL_EMBEDDING_API_KEY = None
        settings.LOCAL_EMBEDDING_MODEL_NAME = None
        settings.LOCAL_EMBEDDING_BASE_URL = None
        settings.USE_LOCAL_EMBEDDINGS = None

    handled, path, _ = edit_ctx.result

    if not handled and save is not None:
        # invalid --save format (unsupported)
        print("Unsupported --save option. Use --save=dotenv[:path].")
    elif path:
        # persisted to a file
        print(f"Removed local embedding environment variables from {path}.")

    if is_openai_configured():
        print(
            ":raised_hands: OpenAI will still be used by default because OPENAI_API_KEY is set."
        )
    else:
        print(
            "The local embeddings model configuration has been removed. No model is currently configured, but you can set one with the CLI or add credentials to .env[.local]."
        )


#############################################
# Gemini Integration ########################
#############################################


@app.command(name="set-gemini")
def set_gemini_model_env(
    model_name: Optional[str] = typer.Option(
        None, "--model-name", help="Gemini Model name"
    ),
    google_api_key: Optional[str] = typer.Option(
        None,
        "--google-api-key",
        help="Google API Key for Gemini",
    ),
    google_cloud_project: Optional[str] = typer.Option(
        None, "--project-id", help="Google Cloud project ID"
    ),
    google_cloud_location: Optional[str] = typer.Option(
        None, "--location", help="Google Cloud location"
    ),
    save: Optional[str] = typer.Option(
        None,
        "--save",
        help="Persist CLI parameters as environment variables in a dotenv file. "
        "Usage: --save=dotenv[:path] (default: .env.local)",
    ),
):
    if not google_api_key and not (
        google_cloud_project and google_cloud_location
    ):
        typer.echo(
            "You must provide either --google-api-key or both --project-id and --location.",
            err=True,
        )
        raise typer.Exit(code=1)

    settings = get_settings()
    with settings.edit(save=save) as edit_ctx:
        edit_ctx.switch_model_provider(ModelKeyValues.USE_GEMINI_MODEL)

        if google_api_key is not None:
            settings.GOOGLE_API_KEY = google_api_key
            settings.GOOGLE_GENAI_USE_VERTEXAI = False
        else:
            settings.GOOGLE_GENAI_USE_VERTEXAI = True
        if google_cloud_project:
            settings.GOOGLE_CLOUD_PROJECT = google_cloud_project
        if google_cloud_location:
            settings.GOOGLE_CLOUD_LOCATION = google_cloud_location
        if model_name:
            settings.GEMINI_MODEL_NAME = model_name

    handled, path, _ = edit_ctx.result

    if not handled and save is not None:
        # invalid --save format (unsupported)
        print("Unsupported --save option. Use --save=dotenv[:path].")
    elif path:
        # persisted to a file
        print(
            f"Saved environment variables to {path} (ensure it's git-ignored)."
        )
    else:
        # updated in-memory & process env only
        print(
            "Settings updated for this session. To persist, use --save=dotenv[:path] "
            "(default .env.local) or set DEEPEVAL_DEFAULT_SAVE=dotenv:.env.local"
        )

    _model_name = (
        model_name if model_name is not None else settings.GEMINI_MODEL_NAME
    )
    if _model_name is not None:
        print(
            f":raising_hands: Congratulations! You're now using Gemini's `{escape(_model_name)}` for all evals that require an LLM."
        )
    else:
        print(
            ":raising_hands: Congratulations! You're now using Gemini's model for all evals that require an LLM."
        )


@app.command(name="unset-gemini")
def unset_gemini_model_env(
    save: Optional[str] = typer.Option(
        None,
        "--save",
        help="Remove only the Gemini related environment variables from a dotenv file. "
        "Usage: --save=dotenv[:path] (default: .env.local)",
    ),
):
    settings = get_settings()
    with settings.edit(save=save) as edit_ctx:
        settings.GOOGLE_API_KEY = None
        settings.GOOGLE_GENAI_USE_VERTEXAI = None
        settings.GOOGLE_CLOUD_PROJECT = None
        settings.GOOGLE_CLOUD_LOCATION = None
        settings.GEMINI_MODEL_NAME = None
        settings.USE_GEMINI_MODEL = None

    handled, path, _ = edit_ctx.result

    if not handled and save is not None:
        # invalid --save format (unsupported)
        print("Unsupported --save option. Use --save=dotenv[:path].")
    elif path:
        # persisted to a file
        print(f"Removed Gemini model environment variables from {path}.")

    if is_openai_configured():
        print(
            ":raised_hands: OpenAI will still be used by default because OPENAI_API_KEY is set."
        )
    else:
        print(
            "The Gemini model configuration has been removed. No model is currently configured, but you can set one with the CLI or add credentials to .env[.local]."
        )


@app.command(name="set-litellm")
def set_litellm_model_env(
    model_name: str = typer.Argument(..., help="Name of the LiteLLM model"),
    api_key: Optional[str] = typer.Option(
        None,
        "--api-key",
        help="API key for the model. Persisted to dotenv if --save is used; never written to the legacy JSON keystore.",
    ),
    api_base: Optional[str] = typer.Option(
        None, "--api-base", help="Base URL for the model API (if required)"
    ),
    save: Optional[str] = typer.Option(
        None,
        "--save",
        help="Persist CLI parameters as environment variables in a dotenv file. "
        "Usage: --save=dotenv[:path] (default: .env.local)",
    ),
):
    settings = get_settings()
    with settings.edit(save=save) as edit_ctx:
        edit_ctx.switch_model_provider(ModelKeyValues.USE_LITELLM)
        settings.LITELLM_MODEL_NAME = model_name
        if api_key is not None:
            settings.LITELLM_API_KEY = api_key
        if api_base is not None:
            settings.LITELLM_API_BASE = api_base

    handled, path, _ = edit_ctx.result

    if not handled and save is not None:
        # invalid --save format (unsupported)
        print("Unsupported --save option. Use --save=dotenv[:path].")
    elif path:
        # persisted to a file
        print(
            f"Saved environment variables to {path} (ensure it's git-ignored)."
        )
    else:
        # updated in-memory & process env only
        print(
            "Settings updated for this session. To persist, use --save=dotenv[:path] "
            "(default .env.local) or set DEEPEVAL_DEFAULT_SAVE=dotenv:.env.local"
        )

    print(
        f":raising_hands: Congratulations! You're now using LiteLLM's `{escape(model_name)}` for all evals that require an LLM."
    )


@app.command(name="unset-litellm")
def unset_litellm_model_env(
    save: Optional[str] = typer.Option(
        None,
        "--save",
        help="Remove only the LiteLLM related environment variables from a dotenv file. "
        "Usage: --save=dotenv[:path] (default: .env.local)",
    ),
):
    settings = get_settings()
    with settings.edit(save=save) as edit_ctx:
        settings.LITELLM_API_KEY = None
        settings.LITELLM_MODEL_NAME = None
        settings.LITELLM_API_BASE = None
        settings.USE_LITELLM = None

    handled, path, _ = edit_ctx.result

    if not handled and save is not None:
        # invalid --save format (unsupported)
        print("Unsupported --save option. Use --save=dotenv[:path].")
    elif path:
        # persisted to a file
        print(f"Removed LiteLLM model environment variables from {path}.")

    if is_openai_configured():
        print(
            ":raised_hands: OpenAI will still be used by default because OPENAI_API_KEY is set."
        )
    else:
        print(
            "The LiteLLM model configuration has been removed. No model is currently configured, but you can set one with the CLI or add credentials to .env[.local]."
        )


if __name__ == "__main__":
    app()



from deepeval.tracing.tracing import (
    Observer,
    current_span_context,
    trace_manager,
)
from deepeval.openai_agents.extractors import *
from deepeval.tracing.context import current_trace_context
from deepeval.tracing.utils import make_json_serializable
from time import perf_counter
from deepeval.tracing.types import TraceSpanStatus

try:
    from agents.tracing import Span, Trace, TracingProcessor
    from agents.tracing.span_data import (
        AgentSpanData,
        CustomSpanData,
        FunctionSpanData,
        GenerationSpanData,
        GuardrailSpanData,
        HandoffSpanData,
        ResponseSpanData,
        SpanData,
    )
    from deepeval.openai_agents.patch import (
        patch_default_agent_run_single_turn,
        patch_default_agent_run_single_turn_streamed,
    )

    openai_agents_available = True
except ImportError:
    openai_agents_available = False


def _check_openai_agents_available():
    if not openai_agents_available:
        raise ImportError(
            "openai-agents is required for this integration. Install it via your package manager"
        )


class DeepEvalTracingProcessor(TracingProcessor):
    def __init__(self) -> None:
        _check_openai_agents_available()
        patch_default_agent_run_single_turn()
        patch_default_agent_run_single_turn_streamed()
        self.span_observers: dict[str, Observer] = {}

    def on_trace_start(self, trace: "Trace") -> None:
        trace_dict = trace.export()
        _trace_uuid = trace_dict.get("id")
        _thread_id = trace_dict.get("group_id")
        _trace_name = trace_dict.get("workflow_name")
        _trace_metadata = trace_dict.get("metadata")

        _trace = trace_manager.start_new_trace(trace_uuid=str(_trace_uuid))
        _trace.thread_id = str(_thread_id)
        _trace.name = str(_trace_name)
        _trace.metadata = make_json_serializable(_trace_metadata)
        current_trace_context.set(_trace)

        trace_manager.add_span(  # adds a dummy root span
            BaseSpan(
                uuid=_trace_uuid,
                trace_uuid=_trace_uuid,
                parent_uuid=None,
                start_time=perf_counter(),
                name=_trace_name,
                status=TraceSpanStatus.IN_PROGRESS,
                children=[],
            )
        )

    def on_trace_end(self, trace: "Trace") -> None:
        trace_dict = trace.export()
        _trace_uuid = trace_dict.get("id")
        _trace_name = trace_dict.get("workflow_name")

        trace_manager.remove_span(_trace_uuid)  # removing the dummy root span
        trace_manager.end_trace(_trace_uuid)
        current_trace_context.set(None)

    def on_span_start(self, span: "Span") -> None:
        if not span.started_at:
            return
        current_span = current_span_context.get()
        if current_span and isinstance(
            current_span, LlmSpan
        ):  # llm span started by
            return

        span_type = self.get_span_kind(span.span_data)
        observer = Observer(span_type=span_type, func_name="NA")
        if span_type == "llm":
            observer.observe_kwargs["model"] = "temporary model"
        observer.update_span_properties = (
            lambda span_type: update_span_properties(span_type, span.span_data)
        )
        self.span_observers[span.span_id] = observer
        observer.__enter__()

    def on_span_end(self, span: "Span") -> None:
        update_trace_properties_from_span_data(
            current_trace_context.get(), span.span_data
        )

        span_type = self.get_span_kind(span.span_data)
        current_span = current_span_context.get()
        if (
            current_span
            and isinstance(current_span, LlmSpan)
            and span_type == "llm"
        ):  # addtional check if the span kind data is llm too
            update_span_properties(current_span, span.span_data)

        observer = self.span_observers.pop(span.span_id, None)
        if observer:
            observer.__exit__(None, None, None)

    def force_flush(self) -> None:
        pass

    def shutdown(self) -> None:
        pass

    def get_span_kind(self, span_data: "SpanData") -> str:
        if isinstance(span_data, AgentSpanData):
            return "agent"
        if isinstance(span_data, FunctionSpanData):
            return "tool"
        if isinstance(span_data, MCPListToolsSpanData):
            return "tool"
        if isinstance(span_data, GenerationSpanData):
            return "llm"
        if isinstance(span_data, ResponseSpanData):
            return "llm"
        if isinstance(span_data, HandoffSpanData):
            return "custom"
        if isinstance(span_data, CustomSpanData):
            return "base"
        if isinstance(span_data, GuardrailSpanData):
            return "base"
        return "base"



from typing import Optional, Tuple, List, Union
from openai import OpenAI, AsyncOpenAI
from openai.types.chat import ParsedChatCompletion
from pydantic import BaseModel
from io import BytesIO
import base64

from deepeval.models.llms.openai_model import (
    model_pricing,
    structured_outputs_models,
)
from deepeval.models import DeepEvalBaseMLLM
from deepeval.models.llms.utils import trim_and_load_json
from deepeval.test_case import MLLMImage
from deepeval.models.utils import parse_model_name
from deepeval.models.retry_policy import (
    create_retry_decorator,
    sdk_retries_for,
)
from deepeval.constants import ProviderSlug as PS


retry_openai = create_retry_decorator(PS.OPENAI)

valid_multimodal_gpt_models = [
    "gpt-4o",
    "gpt-4o-2024-05-13",
    "gpt-4o-2024-08-06",
    "gpt-4o-2024-11-20",
    "gpt-4o-mini",
    "gpt-4o-mini-2024-07-18",
    "gpt-4.1",
    "gpt-4.1-mini",
    "gpt-4.1-nano",
    "o1",
    "o1-preview",
    "o1-2024-12-17",
    "o1-preview-2024-09-12",
    "gpt-4.5-preview-2025-02-27",
    "o4-mini",
]

default_multimodal_gpt_model = "gpt-4.1"

unsupported_log_probs_multimodal_gpt_models = [
    "o1",
    "o1-preview",
    "o1-2024-12-17",
    "o1-preview-2024-09-12",
    "gpt-4.5-preview-2025-02-27",
    "o4-mini",
]


class MultimodalOpenAIModel(DeepEvalBaseMLLM):
    def __init__(
        self,
        model: Optional[str] = None,
        _openai_api_key: Optional[str] = None,
        *args,
        **kwargs,
    ):
        model_name = None
        if isinstance(model, str):
            model_name = parse_model_name(model)
            if model_name not in valid_multimodal_gpt_models:
                raise ValueError(
                    f"Invalid model. Available Multimodal GPT models: {', '.join(model for model in valid_multimodal_gpt_models)}"
                )
        elif model is None:
            model_name = default_multimodal_gpt_model

        self._openai_api_key = _openai_api_key
        self.args = args
        self.kwargs = kwargs

        super().__init__(model_name, *args, **kwargs)

    ###############################################
    # Generate functions
    ###############################################

    @retry_openai
    def generate(
        self,
        multimodal_input: List[Union[str, MLLMImage]],
        schema: Optional[BaseModel] = None,
    ) -> Tuple[str, float]:
        client = OpenAI(api_key=self._openai_api_key)
        prompt = self.generate_prompt(multimodal_input)

        if schema:
            if self.model_name in structured_outputs_models:
                messages = [{"role": "user", "content": prompt}]
                response = client.beta.chat.completions.parse(
                    model=self.model_name,
                    messages=messages,
                    response_format=schema,
                )
                input_tokens = response.usage.prompt_tokens
                output_tokens = response.usage.completion_tokens
                total_cost = self.calculate_cost(input_tokens, output_tokens)
                generated_text = response.choices[0].message.parsed
                return generated_text, total_cost

        completion = client.chat.completions.create(
            model=self.model_name,
            messages=[{"role": "user", "content": prompt}],
        )
        output = completion.choices[0].message.content
        cost = self.calculate_cost(
            completion.usage.prompt_tokens, completion.usage.completion_tokens
        )
        if schema:
            json_output = trim_and_load_json(output)
            return schema.model_validate(json_output), cost
        else:
            return output, cost

    @retry_openai
    async def a_generate(
        self,
        multimodal_input: List[Union[str, MLLMImage]],
        schema: Optional[BaseModel] = None,
    ) -> Tuple[str, float]:
        client = AsyncOpenAI(api_key=self._openai_api_key)
        prompt = self.generate_prompt(multimodal_input)

        if schema:
            if self.model_name in structured_outputs_models:
                messages = [{"role": "user", "content": prompt}]
                response = await client.beta.chat.completions.parse(
                    model=self.model_name,
                    messages=messages,
                    response_format=schema,
                )
                input_tokens = response.usage.prompt_tokens
                output_tokens = response.usage.completion_tokens
                total_cost = self.calculate_cost(input_tokens, output_tokens)
                generated_text = response.choices[0].message.parsed
                return generated_text, total_cost

        completion = await client.chat.completions.create(
            model=self.model_name,
            messages=[{"role": "user", "content": prompt}],
        )
        output = completion.choices[0].message.content
        cost = self.calculate_cost(
            completion.usage.prompt_tokens, completion.usage.completion_tokens
        )
        if schema:
            json_output = trim_and_load_json(output)
            return schema.model_validate(json_output), cost
        else:
            return output, cost

    ###############################################
    # Other generate functions
    ###############################################

    @retry_openai
    def generate_raw_response(
        self,
        multimodal_input: List[Union[str, MLLMImage]],
        top_logprobs: int = 5,
    ) -> Tuple[ParsedChatCompletion, float]:
        client = self._client()
        prompt = self.generate_prompt(multimodal_input)
        messages = [{"role": "user", "content": prompt}]
        completion = client.chat.completions.create(
            model=self.model_name,
            messages=messages,
            logprobs=True,
            top_logprobs=top_logprobs,
        )
        # Cost calculation
        input_tokens = completion.usage.prompt_tokens
        output_tokens = completion.usage.completion_tokens
        cost = self.calculate_cost(input_tokens, output_tokens)
        return completion, cost

    @retry_openai
    async def a_generate_raw_response(
        self,
        multimodal_input: List[Union[str, MLLMImage]],
        top_logprobs: int = 5,
    ) -> Tuple[ParsedChatCompletion, float]:
        client = self._client(async_mode=True)
        prompt = self.generate_prompt(multimodal_input)
        messages = [{"role": "user", "content": prompt}]
        completion = await client.chat.completions.create(
            model=self.model_name,
            messages=messages,
            logprobs=True,
            top_logprobs=top_logprobs,
        )
        # Cost calculation
        input_tokens = completion.usage.prompt_tokens
        output_tokens = completion.usage.completion_tokens
        cost = self.calculate_cost(input_tokens, output_tokens)
        return completion, cost

    ###############################################
    # Utilities
    ###############################################

    def generate_prompt(
        self, multimodal_input: List[Union[str, MLLMImage]] = []
    ):
        prompt = []
        for ele in multimodal_input:
            if isinstance(ele, str):
                prompt.append({"type": "text", "text": ele})
            elif isinstance(ele, MLLMImage):
                if ele.local:
                    import PIL.Image

                    image = PIL.Image.open(ele.url)
                    visual_dict = {
                        "type": "image_url",
                        "image_url": {
                            "url": f"data:image/jpeg;base64,{self.encode_pil_image(image)}"
                        },
                    }
                else:
                    visual_dict = {
                        "type": "image_url",
                        "image_url": {"url": ele.url},
                    }
                prompt.append(visual_dict)
        return prompt

    def calculate_cost(self, input_tokens: int, output_tokens: int) -> float:
        pricing = model_pricing.get(
            self.model_name, model_pricing["gpt-4.1"]
        )  # Default to 'gpt-4.1' if model not found
        input_cost = input_tokens * pricing["input"]
        output_cost = output_tokens * pricing["output"]
        return input_cost + output_cost

    def encode_pil_image(self, pil_image):
        image_buffer = BytesIO()
        if pil_image.mode in ("RGBA", "LA", "P"):
            pil_image = pil_image.convert("RGB")
        pil_image.save(image_buffer, format="JPEG")
        image_bytes = image_buffer.getvalue()
        base64_encoded_image = base64.b64encode(image_bytes).decode("utf-8")
        return base64_encoded_image

    def _client(self, async_mode: bool = False):
        kw = {"api_key": self._openai_api_key}
        if not sdk_retries_for(PS.OPENAI):
            kw["max_retries"] = 0
        Client = AsyncOpenAI if async_mode else OpenAI
        return Client(**kw)

    def get_model_name(self):
        return self.model_name



from pydantic import BaseModel, Field, PrivateAttr
from typing import Optional, Dict, List
from deepeval.test_case import ToolCall, Turn


class Golden(BaseModel):
    input: str
    actual_output: Optional[str] = Field(
        default=None, serialization_alias="actualOutput"
    )
    expected_output: Optional[str] = Field(
        default=None, serialization_alias="expectedOutput"
    )
    context: Optional[List[str]] = Field(default=None)
    retrieval_context: Optional[List[str]] = Field(
        default=None, serialization_alias="retrievalContext"
    )
    additional_metadata: Optional[Dict] = Field(
        default=None, serialization_alias="additionalMetadata"
    )
    comments: Optional[str] = Field(default=None)
    tools_called: Optional[List[ToolCall]] = Field(
        default=None, serialization_alias="toolsCalled"
    )
    expected_tools: Optional[List[ToolCall]] = Field(
        default=None, serialization_alias="expectedTools"
    )
    source_file: Optional[str] = Field(
        default=None, serialization_alias="sourceFile"
    )
    name: Optional[str] = Field(default=None)
    custom_column_key_values: Optional[Dict[str, str]] = Field(
        default=None, serialization_alias="customColumnKeyValues"
    )
    _dataset_rank: Optional[int] = PrivateAttr(default=None)
    _dataset_alias: Optional[str] = PrivateAttr(default=None)
    _dataset_id: Optional[str] = PrivateAttr(default=None)


class ConversationalGolden(BaseModel):
    scenario: str
    expected_outcome: Optional[str] = Field(
        None, serialization_alias="expectedOutcome"
    )
    user_description: Optional[str] = Field(
        None, serialization_alias="userDescription"
    )
    context: Optional[List[str]] = Field(default=None)
    additional_metadata: Optional[Dict] = Field(
        default=None, serialization_alias="additionalMetadata"
    )
    comments: Optional[str] = Field(default=None)
    name: Optional[str] = Field(default=None)
    custom_column_key_values: Optional[Dict[str, str]] = Field(
        default=None, serialization_alias="customColumnKeyValues"
    )
    turns: Optional[List[Turn]] = Field(default=None)
    _dataset_rank: Optional[int] = PrivateAttr(default=None)
    _dataset_alias: Optional[str] = PrivateAttr(default=None)
    _dataset_id: Optional[str] = PrivateAttr(default=None)



from typing import Optional, List, Dict, Any
from contextvars import ContextVar
from contextlib import contextmanager
from dataclasses import dataclass

from .tracing import trace_manager
from .context import current_trace_context, update_current_trace
from deepeval.prompt import Prompt
from deepeval.metrics import BaseMetric
from deepeval.test_case.llm_test_case import ToolCall


@dataclass
class LlmContext:
    prompt: Optional[Prompt] = None
    metrics: Optional[List[BaseMetric]] = None
    metric_collection: Optional[str] = None
    expected_output: Optional[str] = None
    expected_tools: Optional[List[ToolCall]] = None
    context: Optional[List[str]] = None
    retrieval_context: Optional[List[str]] = None


current_llm_context: ContextVar[Optional[LlmContext]] = ContextVar(
    "current_llm_context", default=LlmContext()
)


@contextmanager
def trace(
    prompt: Optional[Prompt] = None,
    llm_metrics: Optional[List[BaseMetric]] = None,
    llm_metric_collection: Optional[str] = None,
    name: Optional[str] = None,
    tags: Optional[List[str]] = None,
    metadata: Optional[Dict[str, Any]] = None,
    user_id: Optional[str] = None,
    thread_id: Optional[str] = None,
    expected_output: Optional[str] = None,
    expected_tools: Optional[List[ToolCall]] = None,
    context: Optional[List[str]] = None,
    retrieval_context: Optional[List[str]] = None,
    trace_metric_collection: Optional[str] = None,
    trace_metrics: Optional[List[BaseMetric]] = None,
):
    current_trace = current_trace_context.get()

    if not current_trace:
        current_trace = trace_manager.start_new_trace()

    if trace_metrics:
        current_trace.metrics = trace_metrics

    if trace_metric_collection:
        current_trace.metric_collection = trace_metric_collection

    current_trace_context.set(current_trace)

    current_llm_context.set(
        LlmContext(
            prompt=prompt,
            metrics=llm_metrics,
            metric_collection=llm_metric_collection,
            expected_output=expected_output,
            expected_tools=expected_tools,
            context=context,
            retrieval_context=retrieval_context,
        )
    )

    # set the current trace attributes
    if name:
        update_current_trace(name=name)
    if tags:
        update_current_trace(tags=tags)
    if metadata:
        update_current_trace(metadata=metadata)
    if user_id:
        update_current_trace(user_id=user_id)
    if thread_id:
        update_current_trace(thread_id=thread_id)

    yield current_trace



from typing import List, Optional, Dict, Union
from tqdm import tqdm

from deepeval.dataset import Golden
from deepeval.test_case import LLMTestCase
from deepeval.metrics import BiasMetric
from deepeval.benchmarks.base_benchmark import (
    DeepEvalBaseBenchmark,
    DeepEvalBaseBenchmarkResult,
)
from deepeval.models import DeepEvalBaseLLM
from deepeval.benchmarks.equity_med_qa.task import EquityMedQATask
from deepeval.benchmarks.equity_med_qa.template import EquityMedQATemplate
from deepeval.telemetry import capture_benchmark_run
from deepeval.metrics.utils import initialize_model


class EquityMedQA(DeepEvalBaseBenchmark):
    def __init__(
        self,
        tasks: List[EquityMedQATask] = None,
        model: Optional[Union[str, DeepEvalBaseLLM]] = None,
        **kwargs,
    ):
        from deepeval.scorer import Scorer
        import pandas as pd

        super().__init__(**kwargs)
        self.tasks: List[EquityMedQATask] = (
            list(EquityMedQATask) if tasks is None else tasks
        )
        self.scorer = Scorer()
        self.predictions: Optional[pd.DataFrame] = None
        self.task_scores: Optional[pd.DataFrame] = None
        self.overall_score: Optional[float] = None
        self.evaluation_model, self.using_native_evaluation_model = (
            initialize_model(model)
        )

    def evaluate(
        self, model: DeepEvalBaseLLM, *args, **kwargs
    ) -> DeepEvalBaseBenchmarkResult:
        import pandas as pd

        with capture_benchmark_run("EquityMedQA", len(self.tasks)):
            overall_correct_predictions = 0
            overall_total_predictions = 0
            predictions_row = []
            scores_row = []

            for task in self.tasks:
                goldens = self.load_benchmark_dataset(task)
                task_correct_predictions = 0
                task_total_predictions = len(goldens)
                overall_total_predictions += len(goldens)

                for golden in tqdm(
                    goldens[:10], desc=f"Processing {task.value}"
                ):
                    prediction, score = self.predict(model, golden).values()
                    if score:
                        task_correct_predictions += 1
                        overall_correct_predictions += 1
                    predictions_row.append(
                        (
                            task.value,
                            golden.input,
                            prediction,
                            golden.expected_output,
                            score,
                        )
                    )

                task_accuracy = (
                    task_correct_predictions / task_total_predictions
                )
                print(
                    f"EquityMedQA Task Accuracy (task={task.value}): {task_accuracy}"
                )
                scores_row.append((task.value, task_accuracy))

            # Calculate overall accuracy
            overall_accuracy = (
                overall_correct_predictions / overall_total_predictions
            )
            print(f"Overall EquityMedQA Accuracy: {overall_accuracy}")

            # Create a DataFrame from task_results_data
            # Columns: 'Task', 'Input', 'Prediction', 'Score'
            self.predictions = pd.DataFrame(
                predictions_row,
                columns=[
                    "Task",
                    "Input",
                    "Prediction",
                    "Expected Output",
                    "Correct",
                ],
            )
            self.task_scores = pd.DataFrame(
                scores_row, columns=["Task", "Score"]
            )
            self.overall_score = overall_accuracy

            return DeepEvalBaseBenchmarkResult(
                overall_accuracy=overall_accuracy
            )

    def predict(self, model: DeepEvalBaseLLM, golden: Golden) -> Dict:
        prediction = model.generate(golden.input)

        # For native models, shouldn't happen but just in case
        if isinstance(prediction, tuple):
            prediction = prediction[0]

        # Define Metric
        metric = BiasMetric(
            model=self.evaluation_model,
            strict_mode=True,
        )
        score = metric.measure(
            LLMTestCase(input=golden.input, actual_output=prediction),
            _show_indicator=False,
            _log_metric_to_confident=False,
        )
        flipped_score = (
            1 - metric.score if metric.score in [0, 1] else metric.score
        )
        return {"prediction": prediction, "score": int(flipped_score)}

    def load_benchmark_dataset(self, task: EquityMedQATask) -> List[Golden]:
        from datasets import load_dataset

        # Load full dataset
        dataset_mapping = {
            EquityMedQATask.EHAI: EquityMedQATask.EHAI.value + "_dataset",
            EquityMedQATask.FBRT_LLM: EquityMedQATask.FBRT_LLM.value
            + "_dataset",
            EquityMedQATask.FBRT_LLM_661_SAMPLED: EquityMedQATask.FBRT_LLM_661_SAMPLED.value
            + "_dataset",
            EquityMedQATask.FBRT_MANUAL: EquityMedQATask.FBRT_MANUAL.value
            + "_dataset",
            EquityMedQATask.MIXED_MMQA_OMAQ: EquityMedQATask.MIXED_MMQA_OMAQ.value
            + "_dataset",
            EquityMedQATask.MULTIMEDQA: EquityMedQATask.MULTIMEDQA.value
            + "_dataset",
            EquityMedQATask.OMAQ: EquityMedQATask.OMAQ.value + "_dataset",
            EquityMedQATask.OMIYE_ET_AL: EquityMedQATask.OMIYE_ET_AL.value
            + "_dataset",
            EquityMedQATask.TRINDS: EquityMedQATask.TRINDS.value + "_dataset",
        }
        dataset_attr = dataset_mapping.get(task)
        if dataset_attr:
            if not hasattr(self, dataset_attr):
                dataset = load_dataset("katielink/EquityMedQA", task.value)
                setattr(self, dataset_attr, dataset)
            else:
                dataset = getattr(self, dataset_attr)

        # Construct test set
        goldens: List[Golden] = []
        for data in dataset["train"]:
            input = EquityMedQATemplate.format_question(data)
            golden = Golden(input=input)
            goldens.append(golden)
        return goldens



from typing import List
from pydantic import BaseModel


class ReasonScore(BaseModel):
    reason: str
    score: float


class RewrittenReason(BaseModel):
    rewritten_reason: str


class Winner(BaseModel):
    winner: str
    reason: str


class Steps(BaseModel):
    steps: List[str]



from typing import Optional

from deepeval.models import DeepEvalBaseLLM


def should_use_batch(model: DeepEvalBaseLLM, batch_size: Optional[int] = None):
    if batch_size is None:
        return False

    if not hasattr(model, "batch_generate"):
        return False

    return True



from pydantic import (
    BaseModel,
    Field,
    PrivateAttr,
    model_validator,
    AliasChoices,
)
from typing import List, Optional, Dict, Literal
from copy import deepcopy
from enum import Enum

from deepeval.test_case import ToolCall
from deepeval.test_case.mcp import (
    MCPServer,
    MCPPromptCall,
    MCPResourceCall,
    MCPToolCall,
    validate_mcp_servers,
)


class TurnParams(Enum):
    ROLE = "role"
    CONTENT = "content"
    SCENARIO = "scenario"
    EXPECTED_OUTCOME = "expected_outcome"
    RETRIEVAL_CONTEXT = "retrieval_context"
    TOOLS_CALLED = "tools_called"
    MCP_TOOLS = "mcp_tools_called"
    MCP_RESOURCES = "mcp_resources_called"
    MCP_PROMPTS = "mcp_prompts_called"


class Turn(BaseModel):
    role: Literal["user", "assistant"]
    content: str
    user_id: Optional[str] = Field(
        default=None, validation_alias=AliasChoices("userId", "user_id")
    )
    retrieval_context: Optional[List[str]] = Field(
        default=None,
        validation_alias=AliasChoices("retrievalContext", "retrieval_context"),
    )
    tools_called: Optional[List[ToolCall]] = Field(
        default=None,
        validation_alias=AliasChoices("toolsCalled", "tools_called"),
    )
    mcp_tools_called: Optional[List[MCPToolCall]] = Field(default=None)
    mcp_resources_called: Optional[List[MCPResourceCall]] = Field(default=None)
    mcp_prompts_called: Optional[List[MCPPromptCall]] = Field(default=None)
    additional_metadata: Optional[Dict] = Field(
        default=None,
        serialization_alias="additionalMetadata",
        validation_alias=AliasChoices(
            "additionalMetadata", "additional_metadata"
        ),
    )
    _mcp_interaction: bool = PrivateAttr(default=False)

    def __repr__(self):
        attrs = [f"role={self.role!r}", f"content={self.content!r}"]
        if self.user_id is not None:
            attrs.append(f"user_id={self.user_id!r}")
        if self.retrieval_context is not None:
            attrs.append(f"retrieval_context={self.retrieval_context!r}")
        if self.tools_called is not None:
            attrs.append(f"tools_called={self.tools_called!r}")
        if self.mcp_tools_called is not None:
            attrs.append(f"mcp_tools_called={self.mcp_tools_called!r}")
        if self.mcp_resources_called is not None:
            attrs.append(f"mcp_resources_called={self.mcp_resources_called!r}")
        if self.mcp_prompts_called is not None:
            attrs.append(f"mcp_prompts_called={self.mcp_prompts_called!r}")
        if self.additional_metadata is not None:
            attrs.append(f"additional_metadata={self.additional_metadata!r}")
        return f"Turn({', '.join(attrs)})"

    @model_validator(mode="before")
    def validate_input(cls, data):
        mcp_tools_called = data.get("mcp_tools_called")
        mcp_prompts_called = data.get("mcp_prompts_called")
        mcp_resources_called = data.get("mcp_resources_called")

        if (
            mcp_tools_called is not None
            or mcp_prompts_called is not None
            or mcp_resources_called is not None
        ):
            from mcp.types import (
                CallToolResult,
                ReadResourceResult,
                GetPromptResult,
            )

            data["_mcp_interaction"] = True
            if mcp_tools_called is not None:
                if not isinstance(mcp_tools_called, list) or not all(
                    isinstance(tool_called, MCPToolCall)
                    and isinstance(tool_called.result, CallToolResult)
                    for tool_called in mcp_tools_called
                ):
                    raise TypeError(
                        "The 'tools_called' must be a list of 'MCPToolCall' with result of type 'CallToolResult' from mcp.types"
                    )

            if mcp_resources_called is not None:
                if not isinstance(mcp_resources_called, list) or not all(
                    isinstance(resource_called, MCPResourceCall)
                    and isinstance(resource_called.result, ReadResourceResult)
                    for resource_called in mcp_resources_called
                ):
                    raise TypeError(
                        "The 'resources_called' must be a list of 'MCPResourceCall' with result of type 'ReadResourceResult' from mcp.types"
                    )

            if mcp_prompts_called is not None:
                if not isinstance(mcp_prompts_called, list) or not all(
                    isinstance(prompt_called, MCPPromptCall)
                    and isinstance(prompt_called.result, GetPromptResult)
                    for prompt_called in mcp_prompts_called
                ):
                    raise TypeError(
                        "The 'prompts_called' must be a list of 'MCPPromptCall' with result of type 'GetPromptResult' from mcp.types"
                    )

        return data


class ConversationalTestCase(BaseModel):
    turns: List[Turn]
    scenario: Optional[str] = Field(default=None)
    context: Optional[List[str]] = Field(default=None)
    name: Optional[str] = Field(default=None)
    user_description: Optional[str] = Field(
        default=None,
        serialization_alias="userDescription",
        validation_alias=AliasChoices("userDescription", "user_description"),
    )
    expected_outcome: Optional[str] = Field(
        default=None,
        serialization_alias="expectedOutcome",
        validation_alias=AliasChoices("expectedOutcome", "expected_outcome"),
    )
    chatbot_role: Optional[str] = Field(
        default=None,
        serialization_alias="chatbotRole",
        validation_alias=AliasChoices("chatbotRole", "chatbot_role"),
    )
    additional_metadata: Optional[Dict] = Field(
        default=None,
        serialization_alias="additionalMetadata",
        validation_alias=AliasChoices(
            "additionalMetadata", "additional_metadata"
        ),
    )
    comments: Optional[str] = Field(default=None)
    tags: Optional[List[str]] = Field(default=None)
    mcp_servers: Optional[List[MCPServer]] = Field(default=None)

    _dataset_rank: Optional[int] = PrivateAttr(default=None)
    _dataset_alias: Optional[str] = PrivateAttr(default=None)
    _dataset_id: Optional[str] = PrivateAttr(default=None)

    @model_validator(mode="before")
    def validate_input(cls, data):
        turns = data.get("turns")
        context = data.get("context")
        mcp_servers = data.get("mcp_servers")

        if len(turns) == 0:
            raise TypeError("'turns' must not be empty")

        # Ensure `context` is None or a list of strings
        if context is not None:
            if not isinstance(context, list) or not all(
                isinstance(item, str) for item in context
            ):
                raise TypeError("'context' must be None or a list of strings")

        if mcp_servers is not None:
            validate_mcp_servers(mcp_servers)

        copied_turns = []
        for turn in turns:
            if isinstance(turn, Turn):
                copied_turns.append(deepcopy(turn))
            elif isinstance(turn, dict):
                try:
                    copied_turns.append(Turn.model_validate(turn))
                except Exception as e:
                    raise TypeError(f"Invalid dict for Turn: {turn} ({e})")
            else:
                raise TypeError(
                    f"'turns' must be a list of Turn or dict, got {type(turn)}"
                )

        data["turns"] = copied_turns

        return data



from typing import List
from pydantic import BaseModel


class ReasonScore(BaseModel):
    reason: str
    score: float


class BestTestCase(BaseModel):
    best_test_case_index: int = None
    best_test_case_id: str = None
    reason: str


class Steps(BaseModel):
    steps: List[str]



import warnings

try:
    from pydantic_ai.agent import Agent

    is_pydantic_ai_installed = True
except:
    is_pydantic_ai_installed = False


class DeepEvalPydanticAIAgent(Agent):

    def __init__(self, *args, **kwargs):
        warnings.warn(
            "instrument_pydantic_ai is deprecated and will be removed in a future version. "
            "Please use the new ConfidentInstrumentationSettings instead. Docs: https://www.confident-ai.com/docs/integrations/third-party/pydantic-ai",
            DeprecationWarning,
            stacklevel=2,
        )

        super().__init__(*args, **kwargs)



import warnings

from typing import Optional, Tuple, Union, Dict
from anthropic import Anthropic, AsyncAnthropic
from pydantic import BaseModel

from deepeval.models import DeepEvalBaseLLM
from deepeval.models.llms.utils import trim_and_load_json
from deepeval.models.retry_policy import (
    create_retry_decorator,
    sdk_retries_for,
)
from deepeval.models.utils import parse_model_name
from deepeval.config.settings import get_settings
from deepeval.constants import ProviderSlug as PS


# consistent retry rules
retry_anthropic = create_retry_decorator(PS.ANTHROPIC)

model_pricing = {
    "claude-opus-4-20250514": {"input": 15.00 / 1e6, "output": 75.00 / 1e6},
    "claude-sonnet-4-20250514": {"input": 3.00 / 1e6, "output": 15.00 / 1e6},
    "claude-3-7-sonnet-latest": {"input": 3.00 / 1e6, "output": 15.00 / 1e6},
    "claude-3-5-haiku-latest": {"input": 0.80 / 1e6, "output": 4.00 / 1e6},
    "claude-3-5-sonnet-latest": {"input": 3.00 / 1e6, "output": 15.00 / 1e6},
    "claude-3-opus-latest": {"input": 15.00 / 1e6, "output": 75.00 / 1e6},
    "claude-3-sonnet-20240229": {"input": 3.00 / 1e6, "output": 15.00 / 1e6},
    "claude-3-haiku-20240307": {"input": 0.25 / 1e6, "output": 1.25 / 1e6},
    "claude-instant-1.2": {"input": 0.80 / 1e6, "output": 2.40 / 1e6},
}


class AnthropicModel(DeepEvalBaseLLM):
    def __init__(
        self,
        model: str = "claude-3-7-sonnet-latest",
        temperature: float = 0,
        _anthropic_api_key: Optional[str] = None,
        generation_kwargs: Optional[Dict] = None,
        **kwargs,
    ):
        model_name = parse_model_name(model)
        self._anthropic_api_key = _anthropic_api_key

        if temperature < 0:
            raise ValueError("Temperature must be >= 0.")
        self.temperature = temperature

        self.kwargs = kwargs
        self.generation_kwargs = generation_kwargs or {}
        super().__init__(model_name)

    ###############################################
    # Generate functions
    ###############################################

    @retry_anthropic
    def generate(
        self, prompt: str, schema: Optional[BaseModel] = None
    ) -> Tuple[Union[str, Dict], float]:
        chat_model = self.load_model()
        message = chat_model.messages.create(
            max_tokens=1024,
            messages=[
                {
                    "role": "user",
                    "content": prompt,
                }
            ],
            model=self.model_name,
            temperature=self.temperature,
            **self.generation_kwargs,
        )
        cost = self.calculate_cost(
            message.usage.input_tokens, message.usage.output_tokens
        )
        if schema is None:
            return message.content[0].text, cost
        else:
            json_output = trim_and_load_json(message.content[0].text)
            return schema.model_validate(json_output), cost

    @retry_anthropic
    async def a_generate(
        self, prompt: str, schema: Optional[BaseModel] = None
    ) -> Tuple[str, float]:
        chat_model = self.load_model(async_mode=True)
        message = await chat_model.messages.create(
            max_tokens=1024,
            messages=[
                {
                    "role": "user",
                    "content": prompt,
                }
            ],
            model=self.model_name,
            temperature=self.temperature,
            **self.generation_kwargs,
        )
        cost = self.calculate_cost(
            message.usage.input_tokens, message.usage.output_tokens
        )
        if schema is None:
            return message.content[0].text, cost
        else:
            json_output = trim_and_load_json(message.content[0].text)

            return schema.model_validate(json_output), cost

    ###############################################
    # Utilities
    ###############################################

    def calculate_cost(self, input_tokens: int, output_tokens: int) -> float:
        pricing = model_pricing.get(self.model_name)

        if pricing is None:
            # Calculate average cost from all known models
            avg_input_cost = sum(
                p["input"] for p in model_pricing.values()
            ) / len(model_pricing)
            avg_output_cost = sum(
                p["output"] for p in model_pricing.values()
            ) / len(model_pricing)
            pricing = {"input": avg_input_cost, "output": avg_output_cost}

            warnings.warn(
                f"[Warning] Pricing not defined for model '{self.model_name}'. "
                "Using average input/output token costs from existing model_pricing."
            )

        input_cost = input_tokens * pricing["input"]
        output_cost = output_tokens * pricing["output"]
        return input_cost + output_cost

    ###############################################
    # Model
    ###############################################

    def load_model(self, async_mode: bool = False):
        if not async_mode:
            return self._build_client(Anthropic)
        return self._build_client(AsyncAnthropic)

    def get_model_name(self):
        return f"{self.model_name}"

    def _client_kwargs(self) -> Dict:
        kwargs = dict(self.kwargs or {})
        # If we are managing retries with Tenacity, force SDK retries off to avoid double retries.
        # if the user opts into SDK retries via DEEPEVAL_SDK_RETRY_PROVIDERS, then honor their max_retries.
        if not sdk_retries_for(PS.ANTHROPIC):
            kwargs["max_retries"] = 0
        return kwargs

    def _build_client(self, cls):
        settings = get_settings()
        kw = dict(
            api_key=settings.ANTHROPIC_API_KEY or self._anthropic_api_key,
            **self._client_kwargs(),
        )
        try:
            return cls(**kw)
        except TypeError as e:
            # in case older SDKs don’t accept max_retries, drop it and retry
            if "max_retries" in str(e):
                kw.pop("max_retries", None)
                return cls(**kw)
            raise



from typing import Optional
from pydantic import BaseModel, Field


class TaskAndOutcome(BaseModel):
    task: str
    outcome: str


class TaskCompletionVerdict(BaseModel):
    verdict: float
    reason: Optional[str] = Field(default=None)
